import uuid
from gettext import gettext as _

from django.core.management import BaseCommand, CommandError

from pulp_rpm.app.models import Package  # noqa
from pulp_rpm.app.models.advisory import UpdateCollection, UpdateRecord  # noqa


class Command(BaseCommand):
    """
    Django management command for repairing RPM metadata in the Pulp database.
    """

    help = _(__doc__)

    def add_arguments(self, parser):
        """Set up arguments."""
        parser.add_argument("issue", help=_("The github issue # of the issue to be fixed."))

    def handle(self, *args, **options):
        """Implement the command."""
        issue = options["issue"]

        if issue == "2460":
            self.repair_2460()
        if issue == "3127":
            self.repair_3127()
        else:
            raise CommandError(_("Unknown issue: '{}'").format(issue))

    def repair_2460(self):
        """Perform data repair for issue #2460."""

        def fix_package(package):
            def fix_requirement(require):
                (name, flags, epoch, version, release, pre) = require
                if "&#38;#38;" in name:
                    fixed_name = name.replace("&#38;#38;", "&")
                    return (fixed_name, flags, epoch, version, release)
                else:
                    return require

            package.requires = [fix_requirement(require) for require in package.provides]

        # This is the only known affected package, we can update this later if we find more.
        packages = Package.objects.filter(name="bpg-algeti-fonts")

        for package in packages:
            fix_package(package)
            package.save()

    def repair_3127(self):
        """Perform data repair for issue #3127."""
        update_collections = UpdateCollection.objects.exclude(name__isnull=False)
        for collection in update_collections:
            collection.name = "collection-autofill-" + uuid.uuid4().hex[:12]
            collection.save()



# Generated by Django 3.2.13 on 2022-07-15 08:42

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0040_rpmalternatecontentsource'),
    ]

    operations = [
        migrations.CreateModel(
            name='ModulemdObsolete',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='rpm_modulemdobsolete', serialize=False, to='core.content')),
                ('modified', models.DateTimeField()),
                ('module_name', models.TextField()),
                ('module_stream', models.TextField()),
                ('message', models.TextField()),
                ('override_previous', models.BooleanField(null=True)),
                ('module_context', models.TextField(null=True)),
                ('eol_date', models.DateTimeField(null=True)),
                ('obsoleted_by_module_name', models.TextField(null=True)),
                ('obsoleted_by_module_stream', models.TextField(null=True)),
                ('snippet', models.TextField()),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('modified', 'module_name', 'module_stream')},
            },
            bases=('core.content',),
        ),
    ]



from django.conf import settings
from django.urls import path

from .viewsets import CopyViewSet, CompsXmlViewSet

if settings.DOMAIN_ENABLED:
    V3_API_ROOT = settings.V3_DOMAIN_API_ROOT_NO_FRONT_SLASH
else:
    V3_API_ROOT = settings.V3_API_ROOT_NO_FRONT_SLASH

urlpatterns = [
    path(f"{V3_API_ROOT}rpm/copy/", CopyViewSet.as_view({"post": "create"})),
    path(f"{V3_API_ROOT}rpm/comps/", CompsXmlViewSet.as_view({"post": "create"})),
]



"""Tests that perform actions over content unit."""

import os
from tempfile import NamedTemporaryFile

import pytest
import requests

from pulpcore.tests.functional.utils import PulpTaskError
from pulp_rpm.tests.functional.constants import (
    BIG_COMPS_XML,
    BIG_CATEGORY,
    BIG_GROUPS,
    BIG_ENVIRONMENTS,
    BIG_LANGPACK,
    RPM_PACKAGEENVIRONMENT_CONTENT_NAME,
    RPM_PACKAGECATEGORY_CONTENT_NAME,
    RPM_PACKAGEGROUP_CONTENT_NAME,
    RPM_PACKAGELANGPACKS_CONTENT_NAME,
    RPM_UNSIGNED_FIXTURE_URL,
    RPM_PACKAGE_FILENAME,
    RPM_WITH_NON_ASCII_URL,
    SMALL_COMPS_XML,
    SMALL_CATEGORY,
    SMALL_ENVIRONMENTS,
    SMALL_GROUPS,
    SMALL_LANGPACK,
)

SMALL_CONTENT = SMALL_GROUPS + SMALL_CATEGORY + SMALL_LANGPACK + SMALL_ENVIRONMENTS
CENTOS8_CONTENT = BIG_GROUPS + BIG_CATEGORY + BIG_LANGPACK + BIG_ENVIRONMENTS


def test_single_request_unit_and_duplicate_unit(
    delete_orphans_pre, rpm_package_api, monitor_task, pulpcore_bindings
):
    """Test single request upload unit.

    1. Upload a unit
    2. Attempt to upload same unit
    """
    # Single unit upload
    file_to_use = os.path.join(RPM_UNSIGNED_FIXTURE_URL, RPM_PACKAGE_FILENAME)

    with NamedTemporaryFile() as file_to_upload:
        file_to_upload.write(requests.get(file_to_use).content)
        upload_attrs = {"file": file_to_upload.name}
        upload = rpm_package_api.create(**upload_attrs)

    content = monitor_task(upload.task).created_resources[0]
    package = rpm_package_api.read(content)
    assert package.location_href == RPM_PACKAGE_FILENAME

    # Duplicate unit
    with NamedTemporaryFile() as file_to_upload:
        file_to_upload.write(requests.get(file_to_use).content)
        upload_attrs = {"file": file_to_upload.name}
        upload = rpm_package_api.create(**upload_attrs)

    try:
        monitor_task(upload.task)
    except PulpTaskError:
        pass
    task_report = pulpcore_bindings.TasksApi.read(upload.task)
    assert task_report.created_resources[0] == package.pulp_href


def test_upload_non_ascii(delete_orphans_pre, rpm_package_api, monitor_task):
    """Test whether one can upload an RPM with non-ascii metadata."""
    packages_count = rpm_package_api.list().count
    with NamedTemporaryFile() as file_to_upload:
        file_to_upload.write(requests.get(RPM_WITH_NON_ASCII_URL).content)
        upload_attrs = {"file": file_to_upload.name}
        upload = rpm_package_api.create(**upload_attrs)

    monitor_task(upload.task)
    new_packages_count = rpm_package_api.list().count
    assert (packages_count + 1) == new_packages_count


@pytest.fixture
def upload_comps_into(rpm_comps_api, monitor_task):
    def _upload_comps_into(file_path, expected_totals, repo_href=None, replace=False):
        data = {}
        if repo_href:
            data["repository"] = repo_href
            data["replace"] = replace
            expected_totals += 1
        response = rpm_comps_api.rpm_comps_upload(file=file_path, **data)
        task = monitor_task(response.task)
        rsrcs = task.created_resources
        assert expected_totals == len(rsrcs)
        return rsrcs

    return _upload_comps_into


def test_upload_comps_xml(delete_orphans_pre, upload_comps_into):
    """Upload a comps.xml and make sure it created comps-Content."""
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, SMALL_CONTENT)
    eval_resources(resources)


def test_upload_same_comps_xml(delete_orphans_pre, upload_comps_into):
    """Upload a comps.xml twice and make sure it doesn't create new objects the second time."""
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        first = upload_comps_into(comps_file.name, SMALL_CONTENT)
        second = upload_comps_into(comps_file.name, SMALL_CONTENT)

    # we return all resources in the comps.xml, even if they already existed
    assert sorted(first) == sorted(second)


def test_upload_diff_comps_xml(delete_orphans_pre, upload_comps_into):
    """Upload two different comps-files and make sure the second shows the new comps-Content."""
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        upload_comps_into(comps_file.name, SMALL_CONTENT)
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(BIG_COMPS_XML)
        comps_file.flush()
        second = upload_comps_into(comps_file.name, CENTOS8_CONTENT)
    eval_resources(second, is_small=False)


def test_upload_comps_xml_into_repo(
    delete_orphans_pre, rpm_repository_factory, rpm_repository_version_api, upload_comps_into
):
    """Upload comps into a repo and see new version created containing the comps-Content."""
    repo = rpm_repository_factory()

    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, SMALL_CONTENT, repo.pulp_href)

    vers = [g for g in resources if "versions" in g]
    assert len(vers) == 1
    vers_resp = rpm_repository_version_api.read(vers[0])
    assert len(vers_resp.content_summary.added) == 3
    assert len(vers_resp.content_summary.present) == 3
    eval_counts(vers_resp.content_summary.added)


def test_upload_comps_xml_into_repo_add(
    delete_orphans_pre, rpm_repository_factory, rpm_repository_version_api, upload_comps_into
):
    """Upload two comps-files into a repo and see the result being additive."""
    repo = rpm_repository_factory()

    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, SMALL_CONTENT, repo.pulp_href)

    vers = [g for g in resources if "versions" in g]
    assert len(vers) == 1
    vers_resp = rpm_repository_version_api.read(vers[0])
    assert len(vers_resp.content_summary.added) == 3
    assert len(vers_resp.content_summary.present) == 3

    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(BIG_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, CENTOS8_CONTENT, repo.pulp_href)
    vers = [g for g in resources if "versions" in g]
    assert len(vers) == 1
    vers_resp = rpm_repository_version_api.read(vers[0])
    assert vers_resp.number == 2

    assert len(vers_resp.content_summary.added) == 3
    assert len(vers_resp.content_summary.present) == 4
    eval_counts(vers_resp.content_summary.added, is_small=False)
    eval_sum_counts(vers_resp.content_summary.present)


def test_upload_comps_xml_into_repo_replace(
    delete_orphans_pre, rpm_repository_factory, rpm_repository_version_api, upload_comps_into
):
    """Upload two comps, see the comps-content from the second replace the existing."""
    repo = rpm_repository_factory()

    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(SMALL_COMPS_XML)
        comps_file.flush()
        upload_comps_into(comps_file.name, SMALL_CONTENT, repo.pulp_href)
    with NamedTemporaryFile("w+") as comps_file:
        comps_file.write(BIG_COMPS_XML)
        comps_file.flush()
        resources = upload_comps_into(comps_file.name, CENTOS8_CONTENT, repo.pulp_href, True)
    vers = [g for g in resources if "versions" in g]
    assert len(vers) == 1
    vers_resp = rpm_repository_version_api.read(vers[0])
    assert vers_resp.number == 2

    assert len(vers_resp.content_summary.added) == 3
    assert len(vers_resp.content_summary.present) == 3
    eval_counts(vers_resp.content_summary.added, is_small=False)


def eval_resources(resources, is_small=True):
    """Eval created_resources counts."""
    groups = [g for g in resources if "packagegroups" in g]
    assert len(groups) == (SMALL_GROUPS if is_small else BIG_GROUPS)

    categories = [g for g in resources if "packagecategories" in g]
    assert len(categories) == (SMALL_CATEGORY if is_small else BIG_CATEGORY)

    langpacks = [g for g in resources if "packagelangpacks" in g]
    assert len(langpacks) == (SMALL_LANGPACK if is_small else BIG_LANGPACK)

    envs = [g for g in resources if "environment" in g]
    assert len(envs) == (SMALL_ENVIRONMENTS if is_small else BIG_ENVIRONMENTS)


def eval_counts(summary, is_small=True):
    """Eval counts in a given summary."""
    for c in summary:
        if "rpm.packagegroup" in c:
            count = SMALL_GROUPS if is_small else BIG_GROUPS
            assert summary["rpm.packagegroup"]["count"] == count
        elif "rpm.packageenvironment" in c:
            count = SMALL_ENVIRONMENTS if is_small else BIG_ENVIRONMENTS
            assert summary["rpm.packageenvironment"]["count"] == count
        elif RPM_PACKAGECATEGORY_CONTENT_NAME in c:
            count = SMALL_CATEGORY if is_small else BIG_CATEGORY
            assert summary[RPM_PACKAGECATEGORY_CONTENT_NAME]["count"] == count
        elif "rpm.packagelangpacks" in c:
            count = SMALL_LANGPACK if is_small else BIG_LANGPACK
            assert summary["rpm.packagelangpacks"]["count"] == count


def eval_sum_counts(summary):
    """In a given summary, counts should be BIG+SMALL."""
    for c in summary:
        if RPM_PACKAGEGROUP_CONTENT_NAME in c:
            count = BIG_GROUPS + SMALL_GROUPS
            assert summary[RPM_PACKAGEGROUP_CONTENT_NAME]["count"] == count
        elif RPM_PACKAGEENVIRONMENT_CONTENT_NAME in c:
            count = BIG_ENVIRONMENTS + SMALL_ENVIRONMENTS
            assert summary[RPM_PACKAGEENVIRONMENT_CONTENT_NAME]["count"] == count
        elif RPM_PACKAGECATEGORY_CONTENT_NAME in c:
            count = BIG_CATEGORY + SMALL_CATEGORY
            assert summary[RPM_PACKAGECATEGORY_CONTENT_NAME]["count"] == count
        elif RPM_PACKAGELANGPACKS_CONTENT_NAME in c:
            count = BIG_LANGPACK + SMALL_LANGPACK
            assert summary[RPM_PACKAGELANGPACKS_CONTENT_NAME]["count"] == count



"""Tests that perform actions over content unit."""

from textwrap import dedent
from urllib.parse import urljoin

import pytest
import requests
from pulpcore.client.pulp_rpm import RpmModulemdDefaults, RpmModulemd

from pulp_rpm.tests.functional.constants import (
    RPM_KICKSTART_FIXTURE_URL,
    RPM_MODULAR_FIXTURE_URL,
    RPM_PACKAGE_FILENAME,
    RPM_PACKAGE_FILENAME2,
    RPM_REPO_METADATA_FIXTURE_URL,
)
from pulp_rpm.tests.functional.utils import gen_rpm_content_attrs
from pulpcore.tests.functional.utils import PulpTaskError


def test_crud_content_unit(
    delete_orphans_pre,
    signed_artifact,
    gen_object_with_cleanup,
    rpm_package_api,
    rpm_repository_api,
    rpm_repository_factory,
    rpm_repository_version_api,
    monitor_task,
):
    """Test creating, reading, updating, and deleting a content unit of package type."""
    # Create content unit
    content_unit_original = {}

    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME)
    response = rpm_package_api.create(**attrs)
    content_unit = rpm_package_api.read(monitor_task(response.task).created_resources[0])
    # rpm package doesn't keep relative_path but the location href
    del attrs["relative_path"]

    content_unit_original.update(content_unit.to_dict())
    for key, val in attrs.items():
        assert content_unit_original[key] == val

    # Read a content unit by its href
    content_unit = rpm_package_api.read(content_unit_original["pulp_href"]).to_dict()
    for key, val in content_unit_original.items():
        assert content_unit[key] == val

    # Read a content unit by its pkg_id
    page = rpm_package_api.list(pkg_id=content_unit_original["pkg_id"])
    assert len(page.results) == 1
    for key, val in content_unit_original.items():
        assert page.results[0].to_dict()[key] == val

    # Attempt to update a content unit using HTTP PATCH
    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME2)
    with pytest.raises(AttributeError) as exc:
        rpm_package_api.partial_update(content_unit_original["pulp_href"], attrs)
    msg = "object has no attribute 'partial_update'"
    assert msg in str(exc)

    # Attempt to update a content unit using HTTP PUT
    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME2)
    with pytest.raises(AttributeError) as exc:
        rpm_package_api.update(content_unit_original["pulp_href"], attrs)
    msg = "object has no attribute 'update'"
    assert msg in str(exc)

    # Attempt to delete a content unit using HTTP DELETE
    with pytest.raises(AttributeError) as exc:
        rpm_package_api.delete(content_unit_original["pulp_href"])
    msg = "object has no attribute 'delete'"
    assert msg in str(exc)

    # Attempt to create duplicate package without specifying a repository
    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME)
    response = rpm_package_api.create(**attrs)
    duplicate = rpm_package_api.read(monitor_task(response.task).created_resources[0])
    assert duplicate.pulp_href == content_unit_original["pulp_href"]

    # Attempt to create duplicate package while specifying a repository
    repo = rpm_repository_factory()
    attrs = gen_rpm_content_attrs(signed_artifact, RPM_PACKAGE_FILENAME)
    attrs["repository"] = repo.pulp_href
    response = rpm_package_api.create(**attrs)
    monitored_response = monitor_task(response.task)

    duplicate = rpm_package_api.read(monitored_response.created_resources[1])
    assert duplicate.pulp_href == content_unit_original["pulp_href"]

    repo = rpm_repository_api.read(repo.pulp_href)
    assert repo.latest_version_href.endswith("/versions/1/")

    version = rpm_repository_version_api.read(repo.latest_version_href)
    assert version.content_summary.present["rpm.package"]["count"] == 1
    assert version.content_summary.added["rpm.package"]["count"] == 1


@pytest.mark.parallel
@pytest.mark.parametrize(
    "url",
    [RPM_MODULAR_FIXTURE_URL, RPM_KICKSTART_FIXTURE_URL, RPM_REPO_METADATA_FIXTURE_URL],
    ids=["MODULAR_FIXTURE_URL", "KICKSTART_FIXTURE_URL", "REPO_METADATA_FIXTURE_URL"],
)
def test_remove_content_unit(url, init_and_sync, rpm_repository_version_api, bindings_cfg):
    """
    Sync a repository and test that content of any type cannot be removed directly.

    - advisory
    - distribution_tree
    - modulemd
    - modulemd_defaults
    - package
    - packagecategory
    - packageenvironment
    - packagegroup
    - packagelangpacks
    - repo metadata
    """
    repo, _ = init_and_sync(url=url, policy="on_demand")

    # Test remove content by types contained in repository.
    version = rpm_repository_version_api.read(repo.latest_version_href)

    # iterate over content filtered by repository versions
    for content_units in version.content_summary.added.values():
        auth = (bindings_cfg.username, bindings_cfg.password)
        url = urljoin(bindings_cfg.host, content_units["href"])
        response = requests.get(url, auth=auth).json()

        # iterate over particular content units and issue delete requests
        for content_unit in response["results"]:
            url = urljoin(bindings_cfg.host, content_unit["pulp_href"])
            resp = requests.delete(url, auth=auth)

            # check that '405' (method not allowed) is returned
            assert resp.status_code == 405


def test_create_modulemd_defaults(monitor_task, gen_object_with_cleanup, rpm_modulemd_defaults_api):
    """
    Create modulemd_defaults with proper unique identifier.

    See: https://github.com/pulp/pulp_rpm/issues/3495
    """
    request_1 = {
        "module": "squid",
        "stream": "4",
        "profiles": '{"4": ["common"]}',
        "snippet": dedent(
            """\
        ---
        document: modulemd-defaults
        version: 1
        data:
          module: squid
          stream: "4"
          profiles:
            4: [common]
        ..."""
        ),
    }

    # Can create
    modulemd_default = gen_object_with_cleanup(
        rpm_modulemd_defaults_api, RpmModulemdDefaults(**request_1)
    )
    assert modulemd_default.module == request_1["module"]

    # Can create variation
    request_2 = request_1.copy()
    request_2["snippet"] = request_2["snippet"].replace("module: squid", 'module: "squid-mod"')
    modulemd_default = gen_object_with_cleanup(
        rpm_modulemd_defaults_api, RpmModulemdDefaults(**request_2)
    )
    assert modulemd_default.module == request_2["module"]

    # Cant create duplicate
    request_3 = request_1.copy()
    request_3["module"] = "squid-mod2"  # not in unique_togheter
    with pytest.raises(PulpTaskError, match="duplicate key value violates unique constraint"):
        modulemd_default = gen_object_with_cleanup(
            rpm_modulemd_defaults_api, RpmModulemdDefaults(**request_3)
        )


def test_create_modulemds(
    monitor_task, gen_object_with_cleanup, rpm_modulemd_api, rpm_package_factory
):
    package = rpm_package_factory()
    request = {
        "name": "foo",
        "stream": "foo",
        "version": "foo",
        "context": "foo",
        "arch": "foo",
        "artifacts": "[]",
        "dependencies": "[]",
        "packages": [package.pulp_href],
        "snippet": "foobar",
        "profiles": "[]",
        "description": "foo",
    }

    # Can upload
    modulemd = gen_object_with_cleanup(rpm_modulemd_api, RpmModulemd(**request))
    assert modulemd.name == request["name"]

    # Cant create duplicate
    with pytest.raises(PulpTaskError, match="duplicate key value violates unique constraint"):
        modulemd = gen_object_with_cleanup(rpm_modulemd_api, RpmModulemd(**request))

    # Can upload variation
    request2 = request.copy()
    request2["snippet"] = "barfoo"
    modulemd = gen_object_with_cleanup(rpm_modulemd_api, RpmModulemd(**request2))
    assert modulemd.name == request2["name"]



"""Tests that sync rpm plugin repositories."""

import pytest
from random import choice

import dictdiffer
import requests
from django.conf import settings
from django.utils.dateparse import parse_datetime

from pulpcore.tests.functional.utils import PulpTaskError
from pulp_smash.pulp3.utils import (
    gen_repo,
    get_added_content_summary,
    get_added_content,
    get_content,
    get_content_summary,
    get_removed_content,
    wget_download_on_host,
)

from pulp_rpm.tests.functional.constants import (
    AMAZON_MIRROR,
    PULP_TYPE_ADVISORY,
    PULP_TYPE_MODULEMD,
    PULP_TYPE_PACKAGE,
    PULP_TYPE_REPOMETADATA,
    REPO_WITH_XML_BASE_URL,
    RPM_ADVISORY_CONTENT_NAME,
    RPM_ADVISORY_COUNT,
    RPM_ADVISORY_DIFFERENT_PKGLIST_URL,
    RPM_ADVISORY_DIFFERENT_REPO_URL,
    RPM_ADVISORY_INCOMPLETE_PKG_LIST_URL,
    RPM_ADVISORY_NO_UPDATED_DATE,
    RPM_ADVISORY_TEST_ADDED_COUNT,
    RPM_ADVISORY_TEST_ID,
    RPM_ADVISORY_TEST_ID_NEW,
    RPM_ADVISORY_TEST_REMOVE_COUNT,
    RPM_ADVISORY_UPDATED_VERSION_URL,
    RPM_CUSTOM_REPO_METADATA_CHANGED_FIXTURE_URL,
    RPM_CUSTOM_REPO_METADATA_FIXTURE_URL,
    RPM_EPEL_MIRROR_URL,
    RPM_COMPLEX_FIXTURE_URL,
    RPM_COMPLEX_PACKAGE_DATA,
    RPM_FIXTURE_SUMMARY,
    RPM_INVALID_FIXTURE_URL,
    RPM_KICKSTART_DATA,
    RPM_KICKSTART_FIXTURE_SUMMARY,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_MD5_REPO_FIXTURE_URL,
    RPM_MIRROR_LIST_BAD_FIXTURE_URL,
    RPM_MIRROR_LIST_GOOD_FIXTURE_URL,
    RPM_MODULES_STATIC_CONTEXT_FIXTURE_URL,
    RPM_MODULAR_FIXTURE_SUMMARY,
    RPM_MODULAR_FIXTURE_URL,
    RPM_MODULAR_STATIC_FIXTURE_SUMMARY,
    RPM_PACKAGE_CONTENT_NAME,
    RPM_PACKAGE_COUNT,
    RPM_REFERENCES_UPDATEINFO_URL,
    RPM_RICH_WEAK_FIXTURE_URL,
    RPM_SHA_FIXTURE_URL,
    RPM_SHA512_FIXTURE_URL,
    RPM_SIGNED_FIXTURE_URL,
    RPM_UNSIGNED_FIXTURE_URL,
    RPM_UPDATED_UPDATEINFO_FIXTURE_URL,
    SRPM_UNSIGNED_FIXTURE_ADVISORY_COUNT,
    SRPM_UNSIGNED_FIXTURE_PACKAGE_COUNT,
    SRPM_UNSIGNED_FIXTURE_URL,
    RPM_MODULEMD_DEFAULTS_DATA,
    RPM_MODULEMD_OBSOLETES_DATA,
    RPM_MODULEMDS_DATA,
    RPM_ZSTD_METADATA_FIXTURE_URL,
)
from pulp_rpm.tests.functional.utils import gen_rpm_remote
from pulp_rpm.tests.functional.utils import set_up_module as setUpModule  # noqa:F401

from pulpcore.client.pulp_rpm import RpmRepositorySyncURL
from pulpcore.client.pulp_rpm.exceptions import ApiException


@pytest.mark.parallel
def test_sync(init_and_sync):
    """Sync repositories with the rpm plugin."""
    # Create a remote (default) and empty repository
    repository, remote = init_and_sync()

    # Assert that it's synced properly
    latest_version_href = repository.latest_version_href
    assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY
    assert get_added_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY

    # Sync the same repository again
    repository, _ = init_and_sync(repository=repository, remote=remote)

    # Assert that the repository has not changed, the latest version stays the same
    assert latest_version_href == repository.latest_version_href
    assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY


@pytest.mark.parallel
def test_sync_zstd(init_and_sync):
    """Test syncing non-gzip metadata."""
    repository, _ = init_and_sync(url=RPM_ZSTD_METADATA_FIXTURE_URL)

    assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY
    assert get_added_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY


@pytest.mark.parallel
def test_sync_local(init_and_sync, tmpdir):
    """Test syncing from the local filesystem."""
    wget_download_on_host(RPM_UNSIGNED_FIXTURE_URL, str(tmpdir))
    init_and_sync(url=f"file://{tmpdir}/rpm-unsigned/")


@pytest.mark.parallel
def test_sync_from_valid_mirror_list_feed(init_and_sync):
    """Sync RPM content from a mirror list feed which contains a valid remote URL."""
    repository, _ = init_and_sync(url=RPM_MIRROR_LIST_GOOD_FIXTURE_URL)

    assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY
    assert get_added_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY


@pytest.mark.parallel
def test_sync_from_valid_mirror_list_feed_with_params(init_and_sync):
    """Sync RPM content from a mirror list feed which contains a valid remote URL."""
    init_and_sync(url=RPM_EPEL_MIRROR_URL)


@pytest.mark.parallel
def test_sync_from_invalid_mirror_list_feed(init_and_sync):
    """Sync RPM content from a mirror list feed which contains an invalid remote URL."""
    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(url=RPM_MIRROR_LIST_BAD_FIXTURE_URL)

    assert "An invalid remote URL was provided" in exc.value.task.to_dict()["error"]["description"]


@pytest.mark.parallel
def test_sync_modular(init_and_sync):
    """Sync RPM modular content."""
    repository, _ = init_and_sync(url=RPM_MODULAR_FIXTURE_URL)

    assert get_content_summary(repository.to_dict()) == RPM_MODULAR_FIXTURE_SUMMARY
    assert get_added_content_summary(repository.to_dict()) == RPM_MODULAR_FIXTURE_SUMMARY


@pytest.mark.parallel
def test_checksum_constraint(init_and_sync):
    """Verify checksum constraint test case.

    Do the following:

    1. Create and sync a repo using the following
       url=RPM_REFERENCES_UPDATEINFO_URL.
    2. Create and sync a secondary repo using the following
       url=RPM_UNSIGNED_FIXTURE_URL.
       Those urls have RPM packages with the same name.
    3. Assert that the task succeed.
    """
    for url in [RPM_REFERENCES_UPDATEINFO_URL, RPM_UNSIGNED_FIXTURE_URL]:
        repository, _ = init_and_sync(url=url)

        assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY
        assert get_added_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY


@pytest.mark.parallel
@pytest.mark.parametrize("policy", ["on_demand", "immediate"])
def test_kickstart(policy, init_and_sync, rpm_content_distribution_trees_api):
    """Sync repositories with the rpm plugin.

    Do the following:

    1. Create a remote.
    2. Sync the remote.
    3. Assert that the correct number of units were added and are present
       in the repo.
    4. Sync the remote one more time.
    5. Assert that repository version is the same the previous one.
    6. Assert that the same number of packages are present.
    """
    repository, remote = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL, policy=policy)

    latest_version_href = repository.latest_version_href
    assert get_content_summary(repository.to_dict()) == RPM_KICKSTART_FIXTURE_SUMMARY
    assert get_added_content_summary(repository.to_dict()) == RPM_KICKSTART_FIXTURE_SUMMARY

    repository, _ = init_and_sync(repository=repository, remote=remote)

    assert get_content_summary(repository.to_dict()) == RPM_KICKSTART_FIXTURE_SUMMARY
    assert latest_version_href == repository.latest_version_href

    distribution_tree = rpm_content_distribution_trees_api.list(
        repository_version=latest_version_href
    ).results[0]
    assert "RHEL" == distribution_tree.release_short


def test_mutated_packages(init_and_sync):
    """Sync two copies of the same packages.

    Make sure we end up with only one copy.

    Do the following:

    1. Sync.
    3. Assert that the content summary matches what is expected.
    4. Create a new remote w/ using fixture containing updated advisory
       (packages with the same NEVRA as the existing package content, but
       different pkgId).
    5. Sync the remote again.
    6. Assert that repository version is different from the previous one
       but has the same content summary.
    7. Assert that the packages have changed since the last sync.
    """
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)

    assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY
    assert get_added_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY

    # Save the copy of the original packages.
    original_packages = {
        (
            content["name"],
            content["epoch"],
            content["version"],
            content["release"],
            content["arch"],
        ): content
        for content in get_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME]
    }

    # Create a remote with a different test fixture with the same NEVRA but
    # different digests.
    repository, _ = init_and_sync(repository=repository, url=RPM_SIGNED_FIXTURE_URL)

    # In case of "duplicates" the most recent one is chosen, so the old
    # package is removed from and the new one is added to a repo version.
    assert (
        len(get_added_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME])
    ) == RPM_PACKAGE_COUNT
    assert (
        len(get_removed_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME])
    ) == RPM_PACKAGE_COUNT

    # Test that the packages have been modified.
    mutated_packages = {
        (
            content["name"],
            content["epoch"],
            content["version"],
            content["release"],
            content["arch"],
        ): content
        for content in get_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME]
    }

    for nevra in original_packages:
        assert original_packages[nevra]["pkgId"] != mutated_packages[nevra]["pkgId"]


def test_sync_diff_checksum_packages(init_and_sync):
    """Sync two fixture content with same NEVRA and different checksum.

    Make sure we end up with the most recently synced content.

    Do the following:

    1. Create two remotes with same content but different checksums.
        Sync the remotes one after the other.
           a. Sync remote with packages with SHA256: ``RPM_UNSIGNED_FIXTURE_URL``.
           b. Sync remote with packages with SHA512: ``RPM_SHA512_FIXTURE_URL``.
    2. Make sure the latest content is only kept.
    """
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL, policy="on_demand")

    repository, _ = init_and_sync(repository=repository, url=RPM_SHA512_FIXTURE_URL)

    added_content = get_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME]
    removed_content = get_removed_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME]

    # In case of "duplicates" the most recent one is chosen, so the old
    # package is removed from and the new one is added to a repo version.
    assert len(added_content) == RPM_PACKAGE_COUNT
    assert len(removed_content) == RPM_PACKAGE_COUNT

    # Verifying whether the packages with first checksum is removed and second
    # is added.
    assert added_content[0]["checksum_type"] == "sha512"
    assert removed_content[0]["checksum_type"] == "sha256"


@pytest.mark.parallel
def test_mutated_advisory_metadata(init_and_sync):
    """Sync two copies of the same Advisory (only description is updated).

    Make sure we end up with only one copy.

    Do the following:

    1. Create a repository and a remote.
    2. Sync the remote.
    3. Assert that the content summary matches what is expected.
    4. Create a new remote w/ using fixture containing updated advisory
       (updaterecords with the ID as the existing updaterecord content, but
       different metadata).
    5. Sync the remote again.
    6. Assert that repository version is different from the previous one
       but has the same content summary.
    7. Assert that the updaterecords have changed since the last sync.
    """
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL, policy="on_demand")

    assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY
    assert get_added_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY

    original_updaterecords = {
        content["id"]: content
        for content in get_content(repository.to_dict())[RPM_ADVISORY_CONTENT_NAME]
    }

    repository, _ = init_and_sync(repository=repository, url=RPM_UPDATED_UPDATEINFO_FIXTURE_URL)

    assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY
    assert len(get_added_content(repository.to_dict())[RPM_ADVISORY_CONTENT_NAME]) == 4
    assert len(get_removed_content(repository.to_dict())[RPM_ADVISORY_CONTENT_NAME]) == 4

    # Test that the updateinfo have been modified.
    mutated_updaterecords = {
        content["id"]: content
        for content in get_content(repository.to_dict())[RPM_ADVISORY_CONTENT_NAME]
    }

    assert mutated_updaterecords != original_updaterecords
    assert mutated_updaterecords[RPM_ADVISORY_TEST_ID_NEW]["description"] == (
        "Updated Gorilla_Erratum and the updated date contains timezone"
    )


@pytest.mark.parallel
def test_optimize(
    init_and_sync,
    gen_object_with_cleanup,
    rpm_repository_api,
    rpm_rpmremote_api,
    monitor_task,
):
    """Tests that sync is skipped when no critical parameters of the sync change.

    Sync is forced if:

    * optimize=False
    * sync URL is different from the last sync
    * sync_policy changes to 'mirror_complete' from something else
    * download_policy of the remote is changed to 'immediate' from something else
    * the repository has been modified since the last sync
    * (NOT tested) repomd revision or repomd checksum change
    """
    repository, remote = init_and_sync(policy="on_demand")

    # sync again, assert optimized
    repository_sync_data = RpmRepositorySyncURL(
        remote=remote.pulp_href, sync_policy="mirror_content_only"
    )
    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert any(report.code == "sync.was_skipped" for report in task.progress_reports)

    repository_sync_data = RpmRepositorySyncURL(remote=remote.pulp_href, optimize=False)
    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert all(report.code != "sync.was_skipped" for report in task.progress_reports)

    # create a new repo version, sync again, assert not optimized
    repository = rpm_repository_api.read(repository.pulp_href)
    content = choice(get_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME])
    response = rpm_repository_api.modify(
        repository.pulp_href, {"remove_content_units": [content["pulp_href"]]}
    )
    monitor_task(response.task)

    repository_sync_data = RpmRepositorySyncURL(
        remote=remote.pulp_href, sync_policy="mirror_content_only"
    )
    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert all(report.code != "sync.was_skipped" for report in task.progress_reports)

    # change download policy to 'immediate', sync again, assert not optimized
    body = {"policy": "immediate"}
    monitor_task(rpm_rpmremote_api.partial_update(remote.pulp_href, body).task)

    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert all(report.code != "sync.was_skipped" for report in task.progress_reports)

    # create new remote with the same URL and download_policy as the first and run a sync task
    new_remote = gen_object_with_cleanup(rpm_rpmremote_api, gen_rpm_remote())
    repository_sync_data = RpmRepositorySyncURL(
        remote=new_remote.pulp_href, sync_policy="mirror_content_only"
    )
    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert any(report.code == "sync.was_skipped" for report in task.progress_reports)

    # change the URL on the new remote, sync again, assert not optimized
    body = {"url": RPM_RICH_WEAK_FIXTURE_URL}
    monitor_task(rpm_rpmremote_api.partial_update(new_remote.pulp_href, body).task)

    repository_sync_data = RpmRepositorySyncURL(
        remote=new_remote.pulp_href, sync_policy="mirror_content_only"
    )
    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert all(report.code != "sync.was_skipped" for report in task.progress_reports)

    # sync again with the new remote, assert optimized
    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert any(report.code == "sync.was_skipped" for report in task.progress_reports)

    # sync again with sync_policy='mirror_complete', assert not optimized, but repository
    # version is unchanged
    first_sync_repo_version = rpm_repository_api.read(repository.pulp_href).latest_version_href
    repository_sync_data = RpmRepositorySyncURL(
        remote=new_remote.pulp_href, sync_policy="mirror_complete"
    )
    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert all(report.code != "sync.was_skipped" for report in task.progress_reports)

    latest_sync_repo_version = rpm_repository_api.read(repository.pulp_href).latest_version_href

    assert first_sync_repo_version == latest_sync_repo_version

    # sync again with sync_policy='mirror_complete', assert optimized
    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    task = monitor_task(sync_response.task)

    assert any(report.code == "sync.was_skipped" for report in task.progress_reports)


@pytest.mark.parallel
def test_sync_advisory_new_version(init_and_sync):
    """Sync a repository and re-sync with newer version of Advisory.

    Test if advisory with same ID and pkglist, but newer version is updated.

    1. Sync rpm-unsigned repository
    2. Re-sync rpm-advisory-updateversion
    3. Check if the newer version advisory was synced
    """
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)

    repository, _ = init_and_sync(repository=repository, url=RPM_ADVISORY_UPDATED_VERSION_URL)

    # check if newer version advisory was added and older removed
    added_advisories = get_added_content(repository.to_dict())[PULP_TYPE_ADVISORY]
    added_advisory = [
        advisory["version"]
        for advisory in added_advisories
        if advisory["id"] == RPM_ADVISORY_TEST_ID
    ]
    removed_advisories = get_removed_content(repository.to_dict())[PULP_TYPE_ADVISORY]
    removed_advisory = [
        advisory["version"]
        for advisory in removed_advisories
        if advisory["id"] == RPM_ADVISORY_TEST_ID
    ]
    assert int(added_advisory[0]) > int(removed_advisory[0])


@pytest.mark.parallel
def test_sync_advisory_old_version(init_and_sync):
    """Sync a repository and re-sync with older version of Advisory.

    Test if advisory with same ID and pkglist, but older version is not updated.

    1. Sync rpm-advisory-updateversion
    2. Re-sync rpm-unsigned repository
    3. Check if the newer (already present) version is preserved
    """
    repository, remote = init_and_sync(url=RPM_ADVISORY_UPDATED_VERSION_URL)
    repository_version_old = repository.latest_version_href

    repository, _ = init_and_sync(repository=repository, url=RPM_UNSIGNED_FIXTURE_URL)
    repository_version_new = repository.latest_version_href

    present_advisories = get_content(repository.to_dict())[PULP_TYPE_ADVISORY]
    advisory_version = [
        advisory["version"]
        for advisory in present_advisories
        if advisory["id"] == RPM_ADVISORY_TEST_ID
    ]

    # check if the newer version is preserved
    assert advisory_version[0] == "2"
    # no new content is present in RPM_UNSIGNED_FIXTURE_URL against
    # RPM_ADVISORY_UPDATED_VERSION_URL so repository latests version should stay the same.
    assert repository_version_old == repository_version_new


@pytest.mark.parallel
def test_sync_merge_advisories(init_and_sync):
    """Sync two advisories with same ID, version and different pkglist.

    Test if two advisories are merged.
    """
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)
    repository, _ = init_and_sync(repository=repository, url=RPM_ADVISORY_DIFFERENT_PKGLIST_URL)

    # check advisories were merged
    added_advisories = get_added_content(repository.to_dict())[PULP_TYPE_ADVISORY]
    added_advisory_pkglist = [
        advisory["pkglist"]
        for advisory in added_advisories
        if advisory["id"] == RPM_ADVISORY_TEST_ID
    ]
    removed_advisories = get_removed_content(repository.to_dict())[PULP_TYPE_ADVISORY]
    removed_advisory_pkglist = [
        advisory["pkglist"]
        for advisory in removed_advisories
        if advisory["id"] == RPM_ADVISORY_TEST_ID
    ]
    added_count = 0
    removed_count = 0
    for collection in added_advisory_pkglist[0]:
        added_count += len(collection["packages"])
    for collection in removed_advisory_pkglist[0]:
        removed_count += len(collection["packages"])

    assert RPM_ADVISORY_TEST_REMOVE_COUNT == removed_count
    assert RPM_ADVISORY_TEST_ADDED_COUNT == added_count


@pytest.mark.parallel
def test_sync_advisory_diff_repo(init_and_sync):
    """Test failure sync advisories.

    If advisory has same id, version but different update_date and
    no packages intersection sync should fail.

    Tested error_msg must be same as we use in pulp_rpm.app.advisory.

    NOTE: If ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION is True, this test
    will fail since the errata-merge will be allowed.
    """
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)
    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(repository=repository, url=RPM_ADVISORY_DIFFERENT_REPO_URL)

    error_msg = (
        "Incoming and existing advisories have the same id but "
        "different timestamps and non-intersecting package lists. "
        "It is likely that they are from two different incompatible remote "
        "repositories. E.g. RHELX-repo and RHELY-debuginfo repo. "
        "Ensure that you are adding content for the compatible repositories. "
        "To allow this behavior, set "
        "ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION = True (q.v.) "
        "in your configuration. Advisory id: {}".format(RPM_ADVISORY_TEST_ID)
    )
    assert error_msg in exc.value.task.to_dict()["error"]["description"]


@pytest.mark.parallel
def test_sync_advisory_proper_subset_pgk_list(init_and_sync):
    """Test success: sync advisories where pkglist is proper-subset of another.

    If update_dates and update_version are the same, pkglist intersection is non-empty
    and a proper-subset of the 'other' pkglist, sync should succeed.
    """
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)
    init_and_sync(repository=repository, url=RPM_ADVISORY_INCOMPLETE_PKG_LIST_URL)


@pytest.mark.parallel
def test_sync_advisory_incomplete_pgk_list(init_and_sync):
    """Test failure sync advisories.

    If update_dates and update_version are the same, pkglist intersection is non-empty
    and not equal to either pkglist sync should fail.

    Tested error_msg must be same as we use in pulp_rpm.app.advisory.
    """
    pytest.skip(reason="Skip until issue #2268 addressed")

    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)

    # create remote with colliding advisory
    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(repository=repository, url=RPM_ADVISORY_INCOMPLETE_PKG_LIST_URL)

    error_msg = (
        "Incoming and existing advisories have the same id and timestamp "
        "but different and intersecting package lists, "
        "and neither package list is a proper subset of the other. "
        "At least one of the advisories is wrong. "
        "Advisory id: {}".format(RPM_ADVISORY_TEST_ID)
    )
    assert error_msg in exc.value.task.to_dict()["error"]["description"]


@pytest.mark.parallel
def test_sync_advisory_no_updated_date(init_and_sync):
    """Test sync advisory with no update.

    1. Sync repository with advisory which has updated_date
    2. Re-sync with repo with same id and version as previous
       but missing updated_date (issued_date should be used instead).
    """
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)
    repository_version = repository.latest_version_href

    repository, _ = init_and_sync(repository=repository, url=RPM_ADVISORY_NO_UPDATED_DATE)
    repository_version_new = repository.latest_version_href

    # TODO: this test isn't ideal because in this case the new advisory with no "updated_date"
    # has an "issued_date" older than the "updated_date" of the already-existing advisory,
    # therefore the new one gets ignored, and nothing happens. A better version of this test
    # would have a newer "issued_date" so that the advisory replaces the old one.
    assert repository_version_new == repository_version

    # added_advisory_date = [
    #     advisory["updated_date"]
    #     for advisory in get_added_content(repo.to_dict())[PULP_TYPE_ADVISORY]
    #     if RPM_ADVISORY_TEST_ID in advisory["id"]
    # ]
    # removed_advisory_date = [
    #     advisory["issued_date"]
    #     for advisory in get_removed_content(repo.to_dict())[PULP_TYPE_ADVISORY]
    #     if RPM_ADVISORY_TEST_ID in advisory["id"]
    # ]

    # self.assertGreater(
    #     parse_datetime(added_advisory_date[0]), parse_datetime(removed_advisory_date[0])
    # )


@pytest.mark.parallel
def test_sync_advisory_updated_update_date(init_and_sync):
    """Test sync advisory with updated update_date."""
    repository, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL)
    repository, _ = init_and_sync(repository=repository, url=RPM_UPDATED_UPDATEINFO_FIXTURE_URL)

    added_advisory_date = [
        advisory["updated_date"]
        for advisory in get_added_content(repository.to_dict())[PULP_TYPE_ADVISORY]
        if RPM_ADVISORY_TEST_ID_NEW in advisory["id"]
    ]
    removed_advisory_date = [
        advisory["updated_date"]
        for advisory in get_removed_content(repository.to_dict())[PULP_TYPE_ADVISORY]
        if RPM_ADVISORY_TEST_ID_NEW in advisory["id"]
    ]

    assert parse_datetime(added_advisory_date[0]) > parse_datetime(removed_advisory_date[0])


@pytest.mark.parallel
def test_sync_advisory_older_update_date(init_and_sync):
    """Test sync advisory with older update_date."""
    repository, _ = init_and_sync(url=RPM_UPDATED_UPDATEINFO_FIXTURE_URL)
    advisory_date = [
        advisory["updated_date"]
        for advisory in get_content(repository.to_dict())[PULP_TYPE_ADVISORY]
        if advisory["id"] == RPM_ADVISORY_TEST_ID
    ]

    repository, _ = init_and_sync(repository, url=RPM_UNSIGNED_FIXTURE_URL)
    advisory_date_new = [
        advisory["updated_date"]
        for advisory in get_content(repository.to_dict())[PULP_TYPE_ADVISORY]
        if advisory["id"] == RPM_ADVISORY_TEST_ID
    ]
    added_advisories = [
        advisory["id"] for advisory in get_added_content(repository.to_dict())[PULP_TYPE_ADVISORY]
    ]

    # check if advisory is preserved and no advisory with same id was added
    assert parse_datetime(advisory_date[0]) == parse_datetime(advisory_date_new[0])
    assert RPM_ADVISORY_TEST_ID not in added_advisories


@pytest.mark.parallel
def test_sync_repo_metadata_change(init_and_sync):
    """Sync RPM modular content."""
    repository, _ = init_and_sync(url=RPM_CUSTOM_REPO_METADATA_FIXTURE_URL)
    repository, _ = init_and_sync(
        repository=repository, url=RPM_CUSTOM_REPO_METADATA_CHANGED_FIXTURE_URL
    )

    # Check if repository was updated with repository metadata
    assert repository.latest_version_href.rstrip("/")[-1] == "2"
    assert PULP_TYPE_REPOMETADATA in get_added_content(repository.to_dict())


@pytest.mark.parallel
def test_sync_modular_static_context(init_and_sync):
    """Sync RPM modular content that includes the new static_context_field."""
    repository, _ = init_and_sync(url=RPM_MODULES_STATIC_CONTEXT_FIXTURE_URL)

    summary = get_content_summary(repository.to_dict())
    added = get_added_content_summary(repository.to_dict())

    modules = get_content(repository.to_dict())[PULP_TYPE_MODULEMD]
    module_static_contexts = [
        (module["name"], module["version"]) for module in modules if module["static_context"]
    ]
    assert len(module_static_contexts) == 2
    assert summary == RPM_MODULAR_STATIC_FIXTURE_SUMMARY
    assert added == RPM_MODULAR_STATIC_FIXTURE_SUMMARY


@pytest.mark.parallel
@pytest.mark.parametrize("sync_policy", ["mirror_content_only", "additive"])
def test_sync_skip_srpm(init_and_sync, sync_policy):
    """In mirror_content_only mode, skip_types is allowed."""
    repository, _ = init_and_sync(
        url=SRPM_UNSIGNED_FIXTURE_URL, skip_types=["srpm"], sync_policy=sync_policy
    )

    present_package_count = len(get_content(repository.to_dict())[PULP_TYPE_PACKAGE])
    present_advisory_count = len(get_content(repository.to_dict())[PULP_TYPE_ADVISORY])
    assert present_package_count == 0
    assert present_advisory_count == SRPM_UNSIGNED_FIXTURE_ADVISORY_COUNT


@pytest.mark.parallel
def test_sha_checksum(init_and_sync):
    """Test that we can sync a repo using SHA as a checksum."""
    init_and_sync(url=RPM_SHA_FIXTURE_URL)


@pytest.mark.skip("TODO: Need a new test fixture")
@pytest.mark.parallel
def test_one_nevra_two_locations_and_checksums(init_and_sync):
    """Sync a repository known to have one nevra, in two locations, with different content.

    While 'odd', this is a real-world occurrence.
    """
    # init_and_sync(url=CENTOS7_OPSTOOLS_URL, policy="on_demand")
    pass


@pytest.mark.timeout(3600)
@pytest.mark.parallel
def test_requires_urlencoded_paths(init_and_sync):
    """Sync a repository known to FAIL when an RPM has non-urlencoded characters in its path.

    See Amazon, java-11-amazon-corretto-javadoc-11.0.8+10-1.amzn2.x86_64.rpm.

    NOTE: testing that this 'works' requires testing against a webserver that does
    whatever-it-is that Amazon's backend is doing. That's why it requires the external repo.
    The rest of the pulp_rpm test-suite is showing us that the code for this fix isn't
    breaking anyone *else*...
    """
    init_and_sync(url=AMAZON_MIRROR, policy="on_demand")


@pytest.mark.parallel
def test_invalid_url(init_and_sync):
    """Sync a repository using a remote url that does not exist."""
    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(url="http://i-am-an-invalid-url.com/invalid/")

    assert exc.value.task.to_dict()["error"]["description"] is not None


@pytest.mark.parallel
def test_invalid_rpm_content(init_and_sync):
    """Sync a repository using an invalid plugin_content repository."""
    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(url=RPM_INVALID_FIXTURE_URL)

    for key in ("missing", "filelists.xml"):
        assert key in exc.value.task.to_dict()["error"]["description"]


@pytest.mark.parallel
def test_sync_metadata_with_unsupported_checksum_type(init_and_sync):
    """
    Sync an RPM repository with an unsupported checksum (md5).

    This test require disallowed 'MD5' checksum type from ALLOWED_CONTENT_CHECKSUMS settings.
    """
    if "md5" in settings.ALLOWED_CONTENT_CHECKSUMS:
        pytest.skip(
            reason="Cannot verify the expected hasher error if the 'MD5' checksum is allowed."
        )

    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(url=RPM_MD5_REPO_FIXTURE_URL)

    assert (
        "does not contain at least one trusted hasher which "
        "is specified in the 'ALLOWED_CONTENT_CHECKSUMS'"
    ) in exc.value.task.to_dict()["error"]["description"]


@pytest.mark.parallel
def test_sync_packages_with_unsupported_checksum_type(init_and_sync):
    """
    Sync an RPM repository with an unsupported checksum (md5) used for packages.

    This test require disallowed 'MD5' checksum type from ALLOWED_CONTENT_CHECKSUMS settings.
    """
    pytest.skip(
        reason=(
            "Needs a repo where an unacceptable checksum is used for packages, but not for metadata"
        )
    )

    if "md5" in settings.ALLOWED_CONTENT_CHECKSUMS:
        pytest.skip(
            reason="Cannot verify the expected hasher error if the 'MD5' checksum is allowed."
        )

    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(url="https://fixtures.com/packages_with_unsupported_checksum")

    error_description = exc.value.task.to_dict()["error"]["description"]
    assert "rpm-with-md5/bear-4.1.noarch.rpm contains forbidden checksum type" in error_description


@pytest.mark.parallel
def test_complete_mirror_with_xml_base_fails(init_and_sync):
    """Test that syncing a repository that uses xml:base in mirror mode fails."""
    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(url=REPO_WITH_XML_BASE_URL, sync_policy="mirror_complete")

    error_description = exc.value.task.to_dict()["error"]["description"]
    assert "features which are incompatible with 'mirror' sync" in error_description


@pytest.mark.parallel
def test_complete_mirror_with_external_location_href_fails(init_and_sync):
    """
    Test that syncing a repository that contains an external location_href fails in mirror mode.

    External location_href refers to a location_href that points outside of the repo,
    e.g. ../../Packages/blah.rpm
    """
    pytest.skip(reason="Needs a repository that links content to a remote source")

    with pytest.raises(PulpTaskError) as exc:
        init_and_sync(
            url="https://fixtures.com/repo_with_external_data", sync_policy="mirror_complete"
        )

    error_description = exc.value.task.to_dict()["error"]["description"]
    assert "features which are incompatible with 'mirror' sync" in error_description


# We can restore this test when we are able to generate repositories on-demand. We just need to
# create a "prestodelta" entry in the repomd.xml, we need not have it actually be a valid one.
@pytest.mark.skip("No DRPM fixture repo.")
@pytest.mark.parallel
def test_complete_mirror_with_delta_metadata_fails(init_and_sync):
    """
    Test that syncing a repository that contains prestodelta metadata fails in mirror mode.

    Otherwise we would be mirroring the metadata without mirroring the DRPM packages.
    """
    with pytest.raises(PulpTaskError) as exc:
        pass
        # init_and_sync(url=DRPM_UNSIGNED_FIXTURE_URL, sync_policy="mirror_complete")

    error_description = exc.value.task.to_dict()["error"]["description"]
    assert "features which are incompatible with 'mirror' sync" in error_description


@pytest.mark.parallel
def test_mirror_and_sync_policy_provided_simultaneously_fails(
    gen_object_with_cleanup,
    rpm_repository_api,
    rpm_rpmremote_api,
):
    """
    Test that syncing fails if both the "mirror" and "sync_policy" params are provided.
    """
    repository = gen_object_with_cleanup(rpm_repository_api, gen_repo())
    remote = gen_object_with_cleanup(
        rpm_rpmremote_api, gen_rpm_remote(url=RPM_UNSIGNED_FIXTURE_URL, policy="on_demand")
    )

    repository_sync_data = RpmRepositorySyncURL(
        remote=remote.pulp_href, sync_policy="mirror_complete", mirror=True
    )

    with pytest.raises(ApiException):
        rpm_repository_api.sync(repository.pulp_href, repository_sync_data)


@pytest.mark.parallel
def test_sync_skip_srpm_fails_mirror_complete(init_and_sync):
    """Test that sync is rejected if both skip_types and mirror_complete are configured."""
    with pytest.raises(ApiException):
        init_and_sync(
            url=SRPM_UNSIGNED_FIXTURE_URL, skip_types=["srpm"], sync_policy="mirror_complete"
        )


@pytest.mark.parallel
def test_core_metadata(init_and_sync, rpm_package_api):
    """Test that the metadata returned by the Pulp API post-sync matches what we expect.

    Do the following:

    1. Sync a repo.
    2. Query package metadata from the API.
    3. Match it against the metadata that we expect to be there.
    """
    repository, _ = init_and_sync(url=RPM_COMPLEX_FIXTURE_URL, policy="on_demand")

    package = rpm_package_api.list(
        name=RPM_COMPLEX_PACKAGE_DATA["name"], repository_version=repository.latest_version_href
    ).results[0]
    package = package.to_dict()

    # sort file and changelog metadata
    package["changelogs"].sort(reverse=True)
    for metadata in [package, RPM_COMPLEX_PACKAGE_DATA]:
        # the list-of-lists can't be sorted easily so we produce a string representation
        files = []
        for f in metadata["files"]:
            files.append(
                "{basename}{filename} type={type}".format(
                    basename=f[1], filename=f[2], type=f[0] or "file"
                )
            )
        metadata["files"] = sorted(files)

    # TODO: figure out how to un-ignore "time_file" without breaking the tests
    diff = dictdiffer.diff(
        package,
        RPM_COMPLEX_PACKAGE_DATA,
        ignore={"time_file", "pulp_created", "pulp_last_updated", "pulp_href", "prn"},
    )
    assert list(diff) == [], list(diff)

    # assert no package is marked modular
    for pkg in get_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME]:
        assert pkg["is_modular"] is False


@pytest.mark.parallel
def test_treeinfo_metadata(init_and_sync, rpm_content_distribution_trees_api):
    """Test that the metadata returned by the Pulp API post-sync matches what we expect.

    Do the following:

    1. Sync a repo.
    2. Query treeinfo metadata from the API.
    3. Match it against the metadata that we expect to be there.
    """
    repository, _ = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL, policy="on_demand")

    distribution_tree = rpm_content_distribution_trees_api.list(
        repository_version=repository.latest_version_href
    ).results[0]
    distribution_tree = distribution_tree.to_dict()
    # delete pulp-specific metadata
    distribution_tree.pop("pulp_href")

    # sort kickstart metadata so that we can compare the dicts properly
    for d in [distribution_tree, RPM_KICKSTART_DATA]:
        d["addons"] = sorted(d["addons"], key=lambda x: x["addon_id"])
        d["images"] = sorted(d["images"], key=lambda x: x["path"])
        d["checksums"] = sorted(d["checksums"], key=lambda x: x["path"])
        d["variants"] = sorted(d["variants"], key=lambda x: x["variant_id"])

    for image in distribution_tree["images"]:
        image.pop("artifact")

    diff = dictdiffer.diff(distribution_tree, RPM_KICKSTART_DATA)
    assert list(diff) == [], list(diff)


def test_modular_metadata(
    init_and_sync,
    rpm_modulemd_api,
    rpm_modulemd_defaults_api,
    rpm_modulemd_obsoletes_api,
    delete_orphans_pre,
):
    """Test that the metadata returned by the Pulp API post-sync matches what we expect.

    Do the following:

    1. Sync a repo.
    2. Query modular metadata from the API.
    3. Match it against the metadata that we expect to be there.
    """
    repository, _ = init_and_sync(url=RPM_MODULAR_FIXTURE_URL, policy="on_demand")

    modules = [
        md.to_dict()
        for md in rpm_modulemd_api.list(repository_version=repository.latest_version_href).results
    ]
    module_defaults = [
        md.to_dict()
        for md in rpm_modulemd_defaults_api.list(
            repository_version=repository.latest_version_href
        ).results
    ]
    module_obsoletes = [
        md.to_dict()
        for md in rpm_modulemd_obsoletes_api.list(
            repository_version=repository.latest_version_href
        ).results
    ]

    def module_key(m):
        return f"{m['name']}-{m['stream']}-{m['version']}-{m['context']}-{m['arch']}"

    def module_default_key(m):
        return f"{m['module']}-{m['stream']}"

    def module_obsolete_key(m):
        return f"{m['module_name']}-{m['module_stream']}"

    modules.sort(key=module_key)
    module_defaults.sort(key=module_default_key)
    module_obsoletes.sort(key=module_obsolete_key)

    RPM_MODULEMDS_DATA.sort(key=module_key)
    RPM_MODULEMD_DEFAULTS_DATA.sort(key=module_default_key)
    RPM_MODULEMD_OBSOLETES_DATA.sort(key=module_obsolete_key)

    for m1, m2 in zip(modules, RPM_MODULEMDS_DATA):
        diff = dictdiffer.diff(
            m1, m2, ignore={"packages", "pulp_created", "pulp_last_updated", "pulp_href", "prn"}
        )
        assert list(diff) == [], list(diff)

    for m1, m2 in zip(module_defaults, RPM_MODULEMD_DEFAULTS_DATA):
        diff = dictdiffer.diff(
            m1, m2, ignore={"pulp_created", "pulp_last_updated", "pulp_href", "prn"}
        )
        assert list(diff) == [], list(diff)

    for m1, m2 in zip(module_obsoletes, RPM_MODULEMD_OBSOLETES_DATA):
        diff = dictdiffer.diff(
            m1, m2, ignore={"pulp_created", "pulp_last_updated", "pulp_href", "prn"}
        )
        assert list(diff) == [], list(diff)

    # assert all package from modular repo is marked as modular
    for pkg in get_content(repository.to_dict())[RPM_PACKAGE_CONTENT_NAME]:
        assert pkg["is_modular"] is True


@pytest.mark.parallel
def test_additive_mode(init_and_sync):
    """Test of additive mode.

    1. Create repository, remote and sync it
    2. Create remote with different set of content
    3. Re-sync and check if new content was added and is present with old one
    """
    repository, remote = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL, policy="on_demand")
    repository, _ = init_and_sync(
        repository=repository,
        url=SRPM_UNSIGNED_FIXTURE_URL,
        policy="on_demand",
        sync_policy="additive",
    )

    present_package_count = len(get_content(repository.to_dict())[PULP_TYPE_PACKAGE])
    present_advisory_count = len(get_content(repository.to_dict())[PULP_TYPE_ADVISORY])

    assert (RPM_PACKAGE_COUNT + SRPM_UNSIGNED_FIXTURE_PACKAGE_COUNT) == present_package_count
    assert (RPM_ADVISORY_COUNT + SRPM_UNSIGNED_FIXTURE_ADVISORY_COUNT) == present_advisory_count


@pytest.mark.parallel
@pytest.mark.parametrize("sync_policy", ["mirror_complete", "mirror_content_only"])
def test_mirror_mode(sync_policy, init_and_sync, rpm_publication_api):
    """Test of mirror mode."""
    repository, remote = init_and_sync(url=SRPM_UNSIGNED_FIXTURE_URL, policy="on_demand")

    assert repository.latest_version_href == f"{repository.pulp_href}versions/1/"
    assert rpm_publication_api.list(repository_version=repository.latest_version_href).count == 0

    repository, _ = init_and_sync(
        repository=repository, url=RPM_SIGNED_FIXTURE_URL, sync_policy=sync_policy
    )

    # check that one publication was created w/ no repository versions
    # and only the new content is present
    assert get_content_summary(repository.to_dict()) == RPM_FIXTURE_SUMMARY
    assert repository.latest_version_href == f"{repository.pulp_href}versions/2/"

    if sync_policy == "mirror_complete":
        created_publications = rpm_publication_api.list(
            repository_version=repository.latest_version_href
        ).count
        assert created_publications == 1


@pytest.mark.parallel
def test_config_repo_mirror_sync(
    rpm_rpmremote_factory,
    rpm_repository_factory,
    rpm_publication_factory,
    rpm_distribution_factory,
    rpm_repository_api,
    rpm_publication_api,
    monitor_task,
):
    """Whether config.repo has the right content for a mirror sync."""

    # TODO: once we have a fixture repository with signed metadata, we should use that
    # and extend this test to see that repo_gpgcheck=1 even if the repository has it disabled
    remote = rpm_rpmremote_factory(url=RPM_UNSIGNED_FIXTURE_URL)
    repository = rpm_repository_factory(
        remote=remote.pulp_href, repo_config={"repo_gpgcheck": 1, "gpgcheck": 1}
    )
    repository_sync_data = RpmRepositorySyncURL(
        remote=remote.pulp_href,
        sync_policy="mirror_complete",
    )

    sync_response = rpm_repository_api.sync(repository.pulp_href, repository_sync_data)
    monitor_task(sync_response.task)
    repository = rpm_repository_api.read(repository.pulp_href)
    publication = rpm_publication_api.list(
        repository_version=repository.latest_version_href
    ).results[0]

    distribution = rpm_distribution_factory(
        publication=publication.pulp_href, generate_repo_config=True
    )

    content = requests.get(f"{distribution.base_url}config.repo").content

    # gpgcheck should follow repo_config
    # repo_gpgcheck should match whether the upstream repo has signed metadata or not
    assert bytes(f"[{distribution.name}]\n", "utf-8") in content
    assert bytes(f"baseurl={distribution.base_url}\n", "utf-8") in content
    assert bytes("gpgcheck=1\n", "utf-8") in content
    assert bytes("repo_gpgcheck=0", "utf-8") in content



from gettext import gettext as _
from collections import defaultdict
from itertools import chain

import hashlib

import createrepo_c as cr
from datetime import datetime

from django.conf import settings
from django.db import (
    IntegrityError,
    transaction,
)
from django.utils.dateparse import parse_datetime

from pulpcore.plugin.models import (
    Content,
    RepositoryContent,
)

from pulp_rpm.app.exceptions import AdvisoryConflict
from pulp_rpm.app.models import (
    UpdateCollectionPackage,
    UpdateRecord,
)
from pulp_rpm.app.shared_utils import is_previous_version


def resolve_advisories(version, previous_version):
    """
    Decide which advisories to add to a repo version and which to remove, and adjust a repo version.

    Advisory can be in 3 different states with relation to a repository version:
     - in-memory and added before this function call, so it's a part of the current incomplete
       repository version only
     - in the db, it's been added in some previous repository version
     - has no relation to any repository version because it's been created in this function as an
       outcome of conflict resolution.

    All 3 states need to be handled differently.
    The in-db ones and newly created are straightforward, just remove/add in a standard way.
    To remove in-memory ones (`content_pks_to_exclude`) from an incomplete repo version,
    one needs to do it directly from RepositoryContent. They've never been a part of a repo
    version, they are also not among the `content_pks_to_add` or `content_pks_to_remove` ones.

    Args:
        version (pulpcore.app.models.RepositoryVersion): current incomplete repository version
        previous_version (pulpcore.app.models.RepositoryVersion):  a version preceding
                                                                   the current incomplete one

    """
    # identify conflicting advisories
    advisory_pulp_type = UpdateRecord.get_pulp_type()
    current_advisories = UpdateRecord.objects.filter(
        pk__in=version.content.filter(pulp_type=advisory_pulp_type)
    )

    # check for any conflict
    unique_advisory_ids = {adv.id for adv in current_advisories}
    if len(current_advisories) == len(unique_advisory_ids):
        # no conflicts
        return

    current_advisories_by_id = defaultdict(list)
    for advisory in current_advisories:
        current_advisories_by_id[advisory.id].append(advisory)

    if previous_version:
        previous_advisories = UpdateRecord.objects.filter(
            pk__in=previous_version.content.filter(pulp_type=advisory_pulp_type)
        )
        previous_advisory_ids = set(previous_advisories.values_list("id", flat=True))

        # diff for querysets works fine but the result is not fully functional queryset,
        # e.g. filtering doesn't work
        added_advisories = current_advisories.difference(previous_advisories)
        added_advisories_by_id = defaultdict(list)
        for advisory in added_advisories:
            added_advisories_by_id[advisory.id].append(advisory)
    else:
        previous_advisory_ids = set()
        added_advisories = current_advisories
        added_advisories_by_id = current_advisories_by_id

    # Conflicts can be in different places and behaviour differs based on that.
    # `in_added`, when conflict happens in the added advisories, this is not allowed and
    # should fail.
    # `added_vs_previous`, a standard conflict between an advisory which is being added and the one
    # in the preceding repo version. This should be resolved according to the heuristics,
    # unless previous repo version has conflicts. In the latter case, the added advisory is picked.
    advisory_id_conflicts = {"in_added": [], "added_vs_previous": []}
    for advisory_id, advisories in current_advisories_by_id.items():
        # we are only interested in conflicts where added advisory is present, we are not trying
        # to fix old conflicts in the existing repo version. There is no real harm in htose,
        # just confusing.
        if len(advisories) > 1 and advisory_id in added_advisories_by_id:
            # if the conflict is in added advisories (2+ advisories with the same id are being
            # added), we need to collect such ids to fail later with
            # a list of all conflicting advisories. No other processing of those is needed.
            if len(added_advisories_by_id[advisory_id]) > 1:
                advisory_id_conflicts["in_added"].append(advisory_id)
            # a standard conflict is detected
            elif advisory_id in previous_advisory_ids:
                advisory_id_conflicts["added_vs_previous"].append(advisory_id)

    content_pks_to_add = set()
    content_pks_to_remove = set()
    content_pks_to_exclude = set()  # exclude from the set of content which is being added

    # Incoming has duplicates - merge till There Is Only One
    for adv_id in advisory_id_conflicts["in_added"]:
        conflict_advisories = current_advisories_by_id[adv_id]
        new_advisory = conflict_advisories[0]
        tmp_advisories = []
        for adv in conflict_advisories[1:]:
            pk_to_add, pk_to_remove, pk_to_exclude = resolve_advisory_conflict(new_advisory, adv)
            if pk_to_add:
                new_advisory = UpdateRecord.objects.filter(pk=pk_to_add[0]).get()
                tmp_advisories.append(new_advisory)
            # If the first duplicate advisory is ended up in to_remove or to_exclude then
            # set the one it was compared against as the one to keep.
            # See https://github.com/pulp/pulp_rpm/pull/3588
            elif new_advisory.pk in pk_to_remove or new_advisory.pk in pk_to_exclude:
                new_advisory = adv
            content_pks_to_exclude.update(pk_to_exclude)
            content_pks_to_remove.update(pk_to_remove)
        added_advisories_by_id[adv_id] = [new_advisory]
        content_pks_to_add.add(new_advisory.pk)
        for tmp_adv in tmp_advisories[:-1]:  # Keep only the last added advisory
            tmp_adv.delete()

    # Incoming has a duplicate advisory-id to previous - resolve
    for advisory_id in advisory_id_conflicts["added_vs_previous"]:
        previous_advisory_qs = previous_advisories.filter(id=advisory_id)
        # there can only be one added advisory at this point otherwise the AdvisoryConflict
        # would have been raised by now
        added_advisory = added_advisories_by_id[advisory_id][0]
        added_advisory.touch()
        if previous_advisory_qs.count() > 1:
            # due to an old bug there could be N advisories with the same id in a repo,
            # this is wrong and there may not be a good way to resolve those, so let's take the
            # new one.
            content_pks_to_add.update([added_advisory.pk])
            content_pks_to_remove.update([adv.pk for adv in previous_advisory_qs])
        else:
            to_add, to_remove, to_exclude = resolve_advisory_conflict(
                previous_advisory_qs.first(), added_advisory
            )
            content_pks_to_add.update(to_add)
            content_pks_to_remove.update(to_remove)
            content_pks_to_exclude.update(to_exclude)

    if content_pks_to_add:
        version.add_content(Content.objects.filter(pk__in=content_pks_to_add))

    if content_pks_to_remove:
        version.remove_content(Content.objects.filter(pk__in=content_pks_to_remove))

    if content_pks_to_exclude:
        RepositoryContent.objects.filter(
            repository=version.repository,
            content_id__in=content_pks_to_exclude,
            version_added=version,
        ).delete()


def resolve_advisory_conflict(previous_advisory, added_advisory):
    """
    Decide which advisory to add to a repo version, create a new one if needed.

    No advisories with the same id can be present in a repo version.

    An existing advisory can be removed from a repo version, a newly added one can stay in a repo
    version, or advisories merge into newly created one which is added to a repo version.
    Merge is done based on criteria described below.

     1. If updated_dates and update_version are the same and pkglist intersection is empty
     (e.g. base repo merged with debuginfo repo) -> new UpdateRecord content unit with combined
     pkglist is created.

     2. If updated_dates or update_version differ and pkglist intersection is non-empty
     (update/re-sync/upload-new case) -> UpdateRecord with newer updated_date or update_version
     is added.

     3. If updated_dates differ and pkglist intersection is empty:
       3.a If pklists differ only IN EVR (ie, name-intersection is Not Empty) -> use-newer
       3.b else -> ERROR CONDITION
         (e.g. base and-debuginfo repos are from different versions, not at same date)

     4. If update_dates and update_version are the same, pkglist intersection is non-empty
     and not a proper subset of to either pkglist - ERROR CONDITION!
     (never-happen case - "something is Terribly Wrong Here")

     Args:
       previous_advisory(pulp_rpm.app.models.UpdateRecord): Advisory which is in a previous repo
                                                            version
       added_advisory(pulp_rpm.app.models.UpdateRecord): Advisory which is being added

     Returns:
       to_add(list): UUIDs of advisories to add to a repo version, can be newly created ones
       to_remove(list): UUIDs of advisories to remove from a repo version
       to_exclude(list): UUIDs of advisories to exclude from the added set of content for a repo
                                  version

    """

    def _datetime_heuristics(in_str):
        # issue- and update-dates can be datetimes, empty, or timetamps. Alas.
        # Try to Do The Right Thing.
        # Return None if we give up
        if not in_str:
            return None

        dt = parse_datetime(in_str)
        if not dt:
            try:
                tstamp = int(in_str)
                dt = datetime.fromtimestamp(tstamp)
            except:  # noqa
                # No idea what this is - give up and return None
                return None
        return dt

    def _do_merge():
        # previous_advisory is used to copy the object and thus the variable refers to a
        # different object after `merge_advisories` call
        previous_advisory_pk = previous_advisory.pk
        merged_advisory = merge_advisories(previous_advisory, added_advisory)
        to_add.append(merged_advisory.pk)
        to_remove.append(previous_advisory_pk)
        to_exclude.append(added_advisory.pk)

    def _name_intersect(prev_pkgs, new_pkgs):
        prev_names = set([x[0] for x in prev_pkgs])
        new_names = set([x[0] for x in new_pkgs])
        return prev_names.intersection(new_names)

    to_add, to_remove, to_exclude = [], [], []

    previous_updated_date = _datetime_heuristics(
        previous_advisory.updated_date or previous_advisory.issued_date
    )
    added_updated_date = _datetime_heuristics(
        added_advisory.updated_date or added_advisory.issued_date
    )
    previous_updated_version = previous_advisory.version
    added_updated_version = added_advisory.version
    previous_pkglist = set(previous_advisory.get_pkglist())
    added_pkglist = set(added_advisory.get_pkglist())

    # Prepare results of conditions for easier use.
    same_dates = previous_updated_date == added_updated_date
    same_version = previous_updated_version == added_updated_version
    pkgs_intersection = previous_pkglist.intersection(added_pkglist)
    names_intersection = _name_intersect(previous_pkglist, added_pkglist)

    if same_dates and same_version and pkgs_intersection:
        if previous_pkglist != added_pkglist:
            # prev and new have different pkg-lists. See if one is a proper-subset of the other;
            # if so, choose the one with the *larger* pkglist. Otherwise, error.
            if previous_pkglist < added_pkglist:
                # new has more pkgs - remove previous
                to_remove.append(previous_advisory.pk)
            elif added_pkglist < previous_pkglist:
                # prev has more pkgs - exclude new
                to_exclude.append(added_advisory.pk)
            else:
                if settings.ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION:
                    _do_merge()
                else:
                    raise AdvisoryConflict(
                        _(
                            "Incoming and existing advisories have the same id and timestamp "
                            "but different and intersecting package lists, "
                            "and neither package list is a proper subset of the other. "
                            "At least one of the advisories is wrong. "
                            "To allow this behavior, set "
                            "ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION = True (q.v.) "
                            "in your configuration. Advisory id: {}"
                        ).format(previous_advisory.id)
                    )
        elif previous_pkglist == added_pkglist:
            # it means some advisory metadata changed without bumping the updated_date or version.
            # There is no way to find out which one is newer, and a user can't fix it,
            # so we are choosing the incoming advisory.
            to_remove.append(previous_advisory.pk)
    elif (not same_dates or (same_dates and not same_version)) and not pkgs_intersection:
        if names_intersection or settings.ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION:
            # Keep "newer" advisory
            if not same_dates:
                if previous_updated_date < added_updated_date:
                    to_remove.append(previous_advisory.pk)
                else:
                    to_exclude.append(added_advisory.pk)
            elif not same_version:
                if is_previous_version(previous_updated_version, added_updated_version):
                    to_remove.append(previous_advisory.pk)
                else:
                    to_exclude.append(added_advisory.pk)
        else:
            raise AdvisoryConflict(
                _(
                    "Incoming and existing advisories have the same id but "
                    "different timestamps and non-intersecting package lists. "
                    "It is likely that they are from two different incompatible remote "
                    "repositories. E.g. RHELX-repo and RHELY-debuginfo repo. "
                    "Ensure that you are adding content for the compatible repositories. "
                    "To allow this behavior, set "
                    "ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION = True (q.v.) "
                    "in your configuration. Advisory id: {}"
                ).format(previous_advisory.id)
            )
    elif not same_dates and pkgs_intersection:
        if previous_updated_date < added_updated_date:
            to_remove.append(previous_advisory.pk)
        else:
            to_exclude.append(added_advisory.pk)
    elif not same_version and pkgs_intersection:
        if is_previous_version(previous_updated_version, added_updated_version):
            to_remove.append(previous_advisory.pk)
        else:
            to_exclude.append(added_advisory.pk)
    elif same_dates and same_version and not pkgs_intersection:
        _do_merge()

    return to_add, to_remove, to_exclude


def _copy_update_collections_for(advisory, collections):
    """
    Deep-copy each UpdateCollection in the_collections, and assign to its new advisory.
    """
    new_collections = []
    with transaction.atomic():
        for collection in collections:
            uc_packages = list(collection.packages.all())
            collection.pk = None
            collection.update_record = advisory
            collection.save()
            new_packages = []
            for a_package in uc_packages:
                a_package.pk = None
                a_package.update_collection = collection
                new_packages.append(a_package)
            UpdateCollectionPackage.objects.bulk_create(new_packages)
            new_collections.append(collection)
    return new_collections


def merge_advisories(previous_advisory, added_advisory):
    """
    Create a new advisory with the merged pkglist.

    Args:
        previous_advisory(pulp_rpm.app.models.UpdateRecord): Advisory which is in a previous repo
                                                             version
        added_advisory(pulp_rpm.app.models.UpdateRecord): Advisory which is being added

    Returns:
        merged_advisory(pulp_rpm.app.models.UpdateRecord): Newly created Advisory with merged
                                                           package list from the other two ones.

    """
    previous_collections = previous_advisory.collections.all()
    added_collections = added_advisory.collections.all()
    references = previous_advisory.references.all()

    # First thing to do is ensure collection-name-uniqueness
    # in the newly-merged advisory.

    # If we've seen a collection-name already, create a new name
    # by appending "_<index>" and rename the collection before
    # merging.
    # "index", in this context, is a monotonically-increasing INTEGER. In the code below, note
    # the lines that insure that, when a new "index" is required, choosing (curr-max+1) is
    # guaranteed.
    # NOTE: keeping the collections separate is DELIBERATE:
    # 1) after the fact, it's clear that an advisory merge happened, and that one set of
    #    packages comes from one place, one from a different one
    # 2) package-grouping can be *relevant* (e.g. base-rpms vs debuginfo-rpms)
    # 3) package-grouping can be *required* (e.g., module-collections must be kept separate)

    def get_new_name(collection, collection_group):
        """
        Generate a new collection-name based on the current one and a list of "indexes".

        Collections are unique-ified by adding "_INTEGER" to their given names.
        Collection-group is keyed by "original" name, with values of "all index-integers seen
          so far".
        New names are coined as "original name + max(index-integers seen so far for that name".
        """
        name_list = collection.name.split("_")
        if name_list[-1].isdecimal():
            name_list.pop(-1)
            orig_name = "_".join(name_list)
        else:
            orig_name = collection.name

        if orig_name in collection_group.keys():
            index = max(collection_group[orig_name]) + 1
        else:
            index = 0
        collection_group.setdefault(orig_name, []).append(index)
        return f"{orig_name}_{index}"

    # Set up the "collection-group" to an initial state - all the current-names, and a list
    # of any "indices" they've been given, so we can know what our current-max-index is
    # before we start renaming collections.
    collection_group = {}
    for collection in chain(previous_collections, added_collections):
        # It's possible for a collection to have NO NAME (not even an empty string)
        if not collection.name:
            collection.name = ""
        name_list = collection.name.split("_")
        if name_list[-1].isdecimal():
            index = name_list.pop(-1)
            orig_name = "_".join(name_list)
            collection_group.setdefault(orig_name, []).append(int(index))

    # Now, prepare for, and merge, the two advisories
    with transaction.atomic():
        # Finalize unique-ified collection names, to use in the event of concurrency-collisions
        # if a collection has no name, we assign it the name "collection".
        names_seen = ["collection"]
        collections_to_merge = set([])
        for collection in chain(previous_collections, added_collections):
            # No packages? ignore
            if collection.packages.count() == 0:
                continue

            # no-name? When merging, ILLEGAL! Give it a name
            if not collection.name:
                collection.name = "collection"

            # If we've seen a collection with "this" name - give it a unique one,
            # and add that name to the "seen" list
            if collection.name in names_seen:
                collection.name = get_new_name(collection, collection_group)
            # Either we've not seen it ever, or we've generated a new "indexed" name
            names_seen.append(collection.name)
            # add the collection to the "set of collections for the merged advisory"
            collections_to_merge.add(collection)

        # Compute digest for the new merged advisory and save it
        merged_advisory_cr = previous_advisory.to_createrepo_c(collections=collections_to_merge)
        merged_digest = hash_update_record(merged_advisory_cr)
        merged_advisory = previous_advisory
        # Need to null both pk (content_ptr_id) and pulp_id here to insure django doesn't
        # find the original advisory instead of making a copy for us
        merged_advisory.pk = None
        merged_advisory.pulp_id = None
        merged_advisory.digest = merged_digest
        try:
            with transaction.atomic():
                merged_advisory.save()
        except IntegrityError:
            merged_advisory = UpdateRecord.objects.get(digest=merged_digest)
            merged_advisory.touch()
        else:
            # For UpdateCollections, make sure we don't re-use the collections for either of the
            # advisories being merged
            _copy_update_collections_for(merged_advisory, collections_to_merge)
            for reference in references:
                # copy reference and add relation for advisory
                reference.pk = None
                reference.save()
                merged_advisory.references.add(reference)

    return merged_advisory


def hash_update_record(update):
    """
    Find the hex digest for an update record xml from creatrepo_c.

    Args:
        update(createrepo_c.UpdateRecord): update record

    Returns:
        str: a hex digest representing the update record

    """
    uinfo = cr.UpdateInfo()
    uinfo.append(update)
    return hashlib.sha256(uinfo.xml_dump().encode("utf-8")).hexdigest()



"""Tests distribution trees."""

import pytest

from pulp_rpm.tests.functional.constants import (
    PULP_TYPE_DISTRIBUTION_TREE,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_DISTRIBUTION_TREE_CHANGED_ADDON_URL,
    RPM_DISTRIBUTION_TREE_CHANGED_MAIN_URL,
    RPM_DISTRIBUTION_TREE_CHANGED_VARIANT_URL,
)


@pytest.mark.parallel
def test_simple_copy_distribution_tree(
    rpm_kickstart_repo_immediate,
    rpm_repository_factory,
    rpm_repository_version_api,
    rpm_copy_api,
    monitor_task,
):
    """Sync repository with a distribution tree."""
    dest_repo = rpm_repository_factory()
    config = [
        {
            "source_repo_version": rpm_kickstart_repo_immediate.latest_version_href,
            "dest_repo": dest_repo.pulp_href,
        }
    ]

    response = rpm_copy_api.copy_content({"config": config, "dependency_solving": True})
    task = monitor_task(response.task)
    assert len(task.created_resources) == 1
    assert task.created_resources[0] == f"{dest_repo.versions_href}1/"

    dest_repo_ver = rpm_repository_version_api.read(task.created_resources[0])
    assert dest_repo_ver.content_summary.added[PULP_TYPE_DISTRIBUTION_TREE]["count"] == 1


@pytest.mark.parallel
def test_dist_tree_copy_as_content(
    rpm_kickstart_repo_immediate,
    rpm_repository_factory,
    rpm_content_distribution_trees_api,
    rpm_copy_api,
    monitor_task,
):
    """Test sync distribution tree repository and copy it."""
    assert rpm_kickstart_repo_immediate.latest_version_href.endswith("/1/")
    source_content = rpm_content_distribution_trees_api.list(
        repository_version=rpm_kickstart_repo_immediate.latest_version_href
    )
    distribution_tree_href = source_content.results[0].pulp_href
    repo_copy = rpm_repository_factory()

    copy_config = [
        {
            "source_repo_version": rpm_kickstart_repo_immediate.latest_version_href,
            "dest_repo": repo_copy.pulp_href,
            "content": [distribution_tree_href],
        }
    ]
    response = rpm_copy_api.copy_content({"config": copy_config, "dependency_solving": True})
    task = monitor_task(response.task)
    assert len(task.created_resources) == 1
    dest_repo_ver = task.created_resources[0]
    assert dest_repo_ver == f"{repo_copy.versions_href}1/"

    dest_content = rpm_content_distribution_trees_api.list(repository_version=dest_repo_ver)
    assert source_content.count == dest_content.count == 1
    assert {p.pulp_href for p in source_content.results} == {
        p.pulp_href for p in dest_content.results
    }


@pytest.mark.parallel
def test_skip_treeinfo(init_and_sync, has_pulp_plugin):
    # Sync repo. Should create only main repo, not subrepos
    _, _, task = init_and_sync(
        url=RPM_KICKSTART_FIXTURE_URL, skip_types=["treeinfo"], return_task=True
    )
    rsrvd = "prn:rpm.rpmrepository" if has_pulp_plugin("core", "3.55") else "/repositories/rpm/rpm/"
    rsrvd_repos = [r for r in task.reserved_resources_record if rsrvd in r]
    assert 1 == len(rsrvd_repos)

    # Sync again, including kstree. Should end up w/ 5 repos reserved.
    _, _, task = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL, return_task=True)
    rsrvd_repos = [r for r in task.reserved_resources_record if rsrvd in r]
    assert 5 == len(rsrvd_repos)


def test_sync_dist_tree_change_addon_repo(init_and_sync, rpm_package_api, delete_orphans_pre):
    """Test changed addon repository."""
    addon_test_pkg_name = "test-srpm02"
    repo, remote = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL)

    # check testing package is not present
    packages = rpm_package_api.list().results
    assert addon_test_pkg_name not in [p.name for p in packages]

    # re-sync w/ new remote & update repo object
    init_and_sync(repository=repo, url=RPM_DISTRIBUTION_TREE_CHANGED_ADDON_URL)

    # check new pacakge is synced to subrepo
    packages = rpm_package_api.list().results
    assert addon_test_pkg_name in [p.name for p in packages]


def test_sync_dist_tree_change_main_repo(init_and_sync, rpm_package_api, delete_orphans_pre):
    """Test changed main repository."""
    main_repo_test_pkg_name = "test-srpm01"
    repo, remote = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL)
    repo_version_num = repo.latest_version_href.rstrip("/")[-1]

    # re-sync w/ new remote & update repo object
    repo, remote = init_and_sync(repository=repo, url=RPM_DISTRIBUTION_TREE_CHANGED_MAIN_URL)
    updated_repo_version_num = repo.latest_version_href.rstrip("/")[-1]

    # Assert new content was added and repo version was increased
    assert repo_version_num < updated_repo_version_num
    packages = rpm_package_api.list().results
    assert main_repo_test_pkg_name in [p.name for p in packages]


def test_sync_dist_tree_change_variant_repo(init_and_sync, rpm_package_api, delete_orphans_pre):
    """Test changed variant repository."""
    variant_test_pkg_name = "test-srpm03"
    repo, remote = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL)

    # check testing package is not present
    packages = rpm_package_api.list().results
    assert variant_test_pkg_name not in [p.name for p in packages]

    # re-sync w/ new remote & update repo object
    init_and_sync(repository=repo, url=RPM_DISTRIBUTION_TREE_CHANGED_VARIANT_URL)

    # check new pacakge is synced to subrepo
    packages = rpm_package_api.list().results
    assert variant_test_pkg_name in [p.name for p in packages]


def test_remove_repo_with_distribution_tree(
    init_and_sync,
    rpm_repository_api,
    rpm_content_distribution_trees_api,
    pulpcore_bindings,
    monitor_task,
):
    """Sync repository with distribution tree and remove the repository."""
    response = pulpcore_bindings.OrphansCleanupApi.cleanup({"orphan_protection_time": 0})
    monitor_task(response.task)

    num_repos_start = rpm_repository_api.list().count
    num_disttrees_start = rpm_content_distribution_trees_api.list().count

    repo, _ = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL)
    task = rpm_repository_api.delete(repo.pulp_href)
    monitor_task(task.task)

    assert rpm_repository_api.list().count == num_repos_start
    # Remove orphans and check if distribution tree was removed.
    response = pulpcore_bindings.OrphansCleanupApi.cleanup({"orphan_protection_time": 0})
    monitor_task(response.task)
    assert rpm_content_distribution_trees_api.list().count == num_disttrees_start



# Generated by Django 2.2.12 on 2020-04-08 17:09

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0006_opensuse_support'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmpublication',
            name='metadata_checksum_type',
            field=models.CharField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], default='sha256', max_length=10),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='rpmpublication',
            name='package_checksum_type',
            field=models.CharField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], default='sha256', max_length=10),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='original_checksum_types',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict),
        ),
    ]



# Generated by Django 4.2.7 on 2023-11-22 18:48

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("rpm", "0055_add_repo_config_field"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="rpmpublication",
            name="sqlite_metadata",
        ),
        migrations.RemoveField(
            model_name="rpmrepository",
            name="sqlite_metadata",
        ),
    ]



from django.db import migrations
import yaml


def fixup_modulemd_artifacts_dependencies(apps, schema_editor):
    """ Set "artifacts" and "dependencies" for any saved post-3.19 back to the 3.18 format.
    """
    Modulemd = apps.get_model("rpm", "Modulemd")

    modules_to_update = []

    for mmd in Modulemd.objects.all().iterator():
        modulemd = yaml.safe_load(mmd.snippet)
        if modulemd:
            mmd.artifacts = modulemd["data"].get("artifacts", {}).get("rpms", [])
            mmd.dependencies = modulemd["data"].get("dependencies", [])
            modules_to_update.append(mmd)
            if len(modules_to_update) >= 100:
                Modulemd.objects.bulk_update(modules_to_update, ["artifacts", "dependencies"])
                modules_to_update.clear()

    Modulemd.objects.bulk_update(modules_to_update, ["artifacts", "dependencies"])


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0047_modulemd_datefield'),
    ]

    operations = [
        migrations.RunPython(fixup_modulemd_artifacts_dependencies),
    ]



from django.db import migrations
import yaml


def fixup_modulemd_profiles(apps, schema_editor):
    """ Set "profiles" for any saved post-3.19 back to the 3.18 format.
    """
    Modulemd = apps.get_model("rpm", "Modulemd")

    modules_to_update = []

    for mmd in Modulemd.objects.all().iterator():
        modulemd = yaml.safe_load(mmd.snippet)
        if modulemd:
            unprocessed_profiles = modulemd["data"].get("profiles", {})
            profiles = {}
            if unprocessed_profiles:
                for name, data in unprocessed_profiles.items():
                    rpms = data.get("rpms")
                    if rpms:
                        profiles[name] = rpms
                mmd.profiles = profiles
                modules_to_update.append(mmd)

            if len(modules_to_update) >= 100:
                Modulemd.objects.bulk_update(modules_to_update, ["profiles"])
                modules_to_update.clear()

    Modulemd.objects.bulk_update(modules_to_update, ["profiles"])


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0048_artifacts_dependencies_fix'),
    ]

    operations = [
        migrations.RunPython(fixup_modulemd_profiles),
    ]



"""Tests that sync rpm plugin repositories."""

import pytest

from pulp_smash.pulp3.utils import (
    get_added_content_summary,
    get_content_summary,
)

from pulp_rpm.tests.functional.constants import (
    PULP_TYPE_PACKAGE,
    RPM_FIXTURE_SUMMARY,
    RPM_KICKSTART_FIXTURE_SUMMARY,
    RPM_MODULAR_FIXTURE_URL,
    RPM_MODULAR_STATIC_FIXTURE_SUMMARY,
    RPM_MODULES_STATIC_CONTEXT_FIXTURE_URL,
)

from pulpcore.client.pulp_rpm import Copy
from pulpcore.client.pulp_rpm.exceptions import ApiException


@pytest.mark.parallel
def test_modular_static_context_copy(
    init_and_sync,
    monitor_task,
    rpm_copy_api,
    rpm_modulemd_api,
    rpm_repository_factory,
    rpm_repository_api,
):
    """Test copying a static_context-using repo to an empty destination."""
    src, _ = init_and_sync(url=RPM_MODULES_STATIC_CONTEXT_FIXTURE_URL)
    dest = rpm_repository_factory()

    data = Copy(
        config=[{"source_repo_version": src.latest_version_href, "dest_repo": dest.pulp_href}],
        dependency_solving=False,
    )
    monitor_task(rpm_copy_api.copy_content(data).task)

    # Check that we have the correct content counts.
    dest = rpm_repository_api.read(dest.pulp_href)
    assert get_content_summary(dest.to_dict()) == RPM_MODULAR_STATIC_FIXTURE_SUMMARY
    assert get_added_content_summary(dest.to_dict()) == RPM_MODULAR_STATIC_FIXTURE_SUMMARY

    modules = rpm_modulemd_api.list(repository_version=dest.latest_version_href).results
    module_static_contexts = [
        (module.name, module.version) for module in modules if module.static_context
    ]
    assert len(module_static_contexts) == 2


class TestCopyWithUnsignedRepoSyncedImmediate:
    def test_basic_copy_all(
        self,
        monitor_task,
        rpm_copy_api,
        rpm_repository_factory,
        rpm_repository_api,
        rpm_unsigned_repo_immediate,
    ):
        """Test copying all the content from one repo to another."""
        src = rpm_unsigned_repo_immediate
        dest = rpm_repository_factory()

        data = Copy(
            config=[{"source_repo_version": src.latest_version_href, "dest_repo": dest.pulp_href}],
            dependency_solving=False,
        )
        monitor_task(rpm_copy_api.copy_content(data).task)

        # Check that we have the correct content counts.
        dest = rpm_repository_api.read(dest.pulp_href)
        assert get_content_summary(dest.to_dict()) == RPM_FIXTURE_SUMMARY
        assert get_added_content_summary(dest.to_dict()) == RPM_FIXTURE_SUMMARY

    def test_copy_none(
        self,
        monitor_task,
        rpm_copy_api,
        rpm_repository_api,
        rpm_repository_factory,
        rpm_unsigned_repo_immediate,
    ):
        """Test copying NO CONTENT from one repo to another."""
        src = rpm_unsigned_repo_immediate
        dest = rpm_repository_factory()

        data = Copy(
            config=[
                {
                    "source_repo_version": src.latest_version_href,
                    "dest_repo": dest.pulp_href,
                    "content": [],
                }
            ],
            dependency_solving=False,
        )
        monitor_task(rpm_copy_api.copy_content(data).task)

        dest = rpm_repository_api.read(dest.pulp_href)
        # Check that no new repo-version was created in dest_repo
        assert "{}versions/0/".format(dest.pulp_href) == dest.latest_version_href

    def test_invalid_config(
        self,
        rpm_copy_api,
        rpm_repository_api,
        rpm_repository_factory,
        rpm_unsigned_repo_immediate,
    ):
        """Test invalid config."""
        src = rpm_unsigned_repo_immediate
        dest = rpm_repository_factory()

        with pytest.raises(ApiException):
            # no list
            data = Copy(
                config={
                    "source_repo_version": src.latest_version_href,
                    "dest_repo": dest.pulp_href,
                },
                dependency_solving=False,
            )
            rpm_copy_api.copy_content(data)

        with pytest.raises(ApiException):
            good = {
                "source_repo_version": src.latest_version_href,
                "dest_repo": dest.pulp_href,
            }
            bad = {"source_repo_version": src.latest_version_href}
            data = Copy(config=[good, bad], dependency_solving=False)
            rpm_copy_api.copy_content(data)

        with pytest.raises(ApiException):
            data = Copy(
                config=[{"source_repo": src.latest_version_href, "dest_repo": dest.pulp_href}],
                dependency_solving=False,
            )
            rpm_copy_api.copy_content(data)

    def test_content(
        self,
        monitor_task,
        rpm_advisory_api,
        rpm_copy_api,
        rpm_repository_api,
        rpm_repository_factory,
        rpm_unsigned_repo_immediate,
    ):
        """Test the content parameter."""
        src = rpm_unsigned_repo_immediate

        content = rpm_advisory_api.list(repository_version=src.latest_version_href).results
        content_to_copy = (content[0].pulp_href, content[1].pulp_href)

        dest = rpm_repository_factory()

        data = Copy(
            config=[
                {
                    "source_repo_version": src.latest_version_href,
                    "dest_repo": dest.pulp_href,
                    "content": content_to_copy,
                }
            ],
            dependency_solving=False,
        )
        monitor_task(rpm_copy_api.copy_content(data).task)

        dest = rpm_repository_api.read(dest.pulp_href)
        dc = rpm_advisory_api.list(repository_version=dest.latest_version_href)
        dest_content = [c.pulp_href for c in dc.results]

        assert sorted(content_to_copy) == sorted(dest_content)

    def test_all_content_recursive(
        self,
        monitor_task,
        rpm_advisory_api,
        rpm_copy_api,
        rpm_package_api,
        rpm_repository_factory,
        rpm_repository_api,
        rpm_unsigned_repo_immediate,
    ):
        """Test requesting all-rpm-update-content/recursive (see #6519)."""
        src = rpm_unsigned_repo_immediate
        dest = rpm_repository_factory()

        advisory_content = rpm_advisory_api.list(repository_version=src.latest_version_href)
        advisories_to_copy = [rslt.pulp_href for rslt in advisory_content.results]

        rpm_content = rpm_package_api.list(repository_version=src.latest_version_href)
        rpms_to_copy = [rslt.pulp_href for rslt in rpm_content.results]

        content_to_copy = set()
        content_to_copy.update(advisories_to_copy)
        content_to_copy.update(rpms_to_copy)

        data = Copy(
            config=[
                {
                    "source_repo_version": src.latest_version_href,
                    "dest_repo": dest.pulp_href,
                    "content": list(content_to_copy),
                }
            ],
            dependency_solving=True,
        )
        monitor_task(rpm_copy_api.copy_content(data).task)

        dest = rpm_repository_api.read(dest.pulp_href)

        # check advisories copied
        dc = rpm_advisory_api.list(repository_version=dest.latest_version_href)
        dest_content = [c.pulp_href for c in dc.results]
        assert sorted(advisories_to_copy) == sorted(dest_content)

        # check rpms copied
        dc = rpm_package_api.list(repository_version=dest.latest_version_href)
        dest_content = [c.pulp_href for c in dc.results]
        assert sorted(rpms_to_copy) == sorted(dest_content)

    def test_strict_copy_package_to_empty_repo(
        self,
        monitor_task,
        rpm_copy_api,
        rpm_package_api,
        rpm_repository_api,
        rpm_repository_factory,
        rpm_unsigned_repo_immediate,
    ):
        """Test copy package and its dependencies to empty repository.

        - Create repository and populate it
        - Create empty repository
        - Use 'copy' to copy 'whale' package with dependencies
        - assert package and its dependencies were copied
        """
        empty_repo = rpm_repository_factory()
        repo = rpm_unsigned_repo_immediate

        packages = rpm_package_api.list(repository_version=repo.latest_version_href, name="whale")
        package_to_copy = [packages.results[0].pulp_href]

        data = Copy(
            config=[
                {
                    "source_repo_version": repo.latest_version_href,
                    "dest_repo": empty_repo.pulp_href,
                    "content": package_to_copy,
                }
            ],
            dependency_solving=True,
        )
        monitor_task(rpm_copy_api.copy_content(data).task)

        empty_repo = rpm_repository_api.read(empty_repo.pulp_href)
        packages = rpm_package_api.list(repository_version=empty_repo.latest_version_href).results
        packages = [package.name for package in packages]

        # assert that only 3 packages are copied (original package with its two dependencies)
        assert len(packages) == 3
        # assert dependencies package names
        for dependency in ["shark", "stork"]:
            assert dependency in packages

    def test_strict_copy_packagecategory_to_empty_repo(
        self,
        monitor_task,
        rpm_copy_api,
        rpm_package_api,
        rpm_package_category_api,
        rpm_package_groups_api,
        rpm_repository_api,
        rpm_repository_factory,
        rpm_unsigned_repo_immediate,
    ):
        """Test copy package and its dependencies to empty repository.

        - Create repository and populate it
        - Create empty destination repository
        - Use 'copy' to copy packagecategory recursively
        - assert packagecategory and its dependencies were copied
        """
        dest_repo = rpm_repository_factory()
        repo = rpm_unsigned_repo_immediate

        package_categories = rpm_package_category_api.list(
            repository_version=repo.latest_version_href
        )
        package_category_to_copy = [package_categories.results[0].pulp_href]
        # repository content counts
        repo_packagecategories_count = package_categories.count
        repo_packagegroups_count = rpm_package_groups_api.list(
            repository_version=repo.latest_version_href
        ).count

        # do the copy
        data = Copy(
            config=[
                {
                    "source_repo_version": repo.latest_version_href,
                    "dest_repo": dest_repo.pulp_href,
                    "content": package_category_to_copy,
                }
            ],
            dependency_solving=True,
        )
        monitor_task(rpm_copy_api.copy_content(data).task)

        # copied repository content counts
        dest_repo = rpm_repository_api.read(dest_repo.pulp_href)
        dest_repo_packages = rpm_package_api.list(repository_version=dest_repo.latest_version_href)
        dest_repo_packages_count = dest_repo_packages.count
        dest_repo_packagecategories_count = rpm_package_category_api.list(
            repository_version=dest_repo.latest_version_href
        ).count
        dest_repo_packagegroups_count = rpm_package_groups_api.list(
            repository_version=dest_repo.latest_version_href
        ).count

        # assert that all dependencies were copied
        assert repo_packagecategories_count == dest_repo_packagecategories_count
        assert repo_packagegroups_count == dest_repo_packagegroups_count
        # Not all packages in repository are dependecies,
        # only packagegroups packages and its dependencies
        assert dest_repo_packages_count == 30
        # Assert only one latest version of 'duck' pacakge was copied
        copied_duck_pkg = [
            duck_pkg.version for duck_pkg in dest_repo_packages.results if duck_pkg.name == "duck"
        ]
        assert copied_duck_pkg == ["0.8"]

    def test_strict_copy_package_to_existing_repo(
        self,
        init_and_sync,
        monitor_task,
        rpm_copy_api,
        rpm_package_api,
        rpm_repository_api,
        rpm_repository_version_api,
        rpm_unsigned_repo_immediate,
    ):
        """Test copy package and its dependencies to empty repository.

        - Create repository and populate it
        - Create second repository with package fulfilling test package dependency
        - Use 'copy' to copy 'whale' package with dependencies
        - assert package and its missing dependencies were copied
        """
        # prepare final_repo - copy to repository
        final_repo, _ = init_and_sync()

        # prepare repository - copy from repository
        repo = rpm_unsigned_repo_immediate

        # remove test package and one dependency package from final repository
        data = {
            "remove_content_units": [
                pkg.pulp_href
                for pkg in rpm_package_api.list(
                    repository_version=final_repo.latest_version_href
                ).results
                if pkg.name in ("shark", "whale")
            ]
        }
        monitor_task(rpm_repository_api.modify(final_repo.pulp_href, data).task)

        final_repo = rpm_repository_api.read(final_repo.pulp_href)

        # get package to copy
        packages = rpm_package_api.list(repository_version=repo.latest_version_href, name="whale")
        package_to_copy = [packages.results[0].pulp_href]

        data = Copy(
            config=[
                {
                    "source_repo_version": repo.latest_version_href,
                    "dest_repo": final_repo.pulp_href,
                    "content": package_to_copy,
                }
            ],
            dependency_solving=True,
        )
        copy_response = monitor_task(rpm_copy_api.copy_content(data).task)
        repository_version = rpm_repository_version_api.read(copy_response.created_resources[0])

        # check only two packages was copied, original package to copy and only one
        # of its dependency as one is already present
        content_summary = repository_version.to_dict()["content_summary"]
        assert content_summary["added"][PULP_TYPE_PACKAGE]["count"] == 2


class TestCopyWithKickstartRepoSyncedImmediate:
    def test_kickstart_content(
        self,
        monitor_task,
        rpm_copy_api,
        rpm_content_distribution_trees_api,
        rpm_kickstart_repo_immediate,
        rpm_repository_api,
        rpm_repository_factory,
    ):
        """Test the content parameter."""
        src = rpm_kickstart_repo_immediate
        dest = rpm_repository_factory()

        content = rpm_content_distribution_trees_api.list(
            repository_version=src.latest_version_href
        )
        content_to_copy = [content.results[0].pulp_href]
        data = Copy(
            config=[
                {
                    "source_repo_version": src.latest_version_href,
                    "dest_repo": dest.pulp_href,
                    "content": content_to_copy,
                }
            ],
            dependency_solving=False,
        )
        monitor_task(rpm_copy_api.copy_content(data).task)

        dest = rpm_repository_api.read(dest.pulp_href)
        content = rpm_content_distribution_trees_api.list(
            repository_version=dest.latest_version_href
        )
        dest_content = [c.pulp_href for c in content.results]

        assert content_to_copy == dest_content

    def test_kickstart_copy_all(
        self,
        monitor_task,
        rpm_copy_api,
        rpm_kickstart_repo_immediate,
        rpm_repository_api,
        rpm_repository_factory,
    ):
        """Test copying all the content from one repo to another."""
        src = rpm_kickstart_repo_immediate
        dest = rpm_repository_factory()

        data = Copy(
            config=[{"source_repo_version": src.latest_version_href, "dest_repo": dest.pulp_href}],
            dependency_solving=False,
        )
        monitor_task(rpm_copy_api.copy_content(data).task)

        # Check that we have the correct content counts.
        dest = rpm_repository_api.read(dest.pulp_href)
        assert get_content_summary(dest.to_dict()) == RPM_KICKSTART_FIXTURE_SUMMARY
        assert get_added_content_summary(dest.to_dict()) == RPM_KICKSTART_FIXTURE_SUMMARY


def test_strict_copy_module_to_empty_repo(
    monitor_task,
    rpm_copy_api,
    rpm_modulemd_api,
    rpm_repository_api,
    rpm_repository_factory,
    rpm_modular_repo_on_demand,
):
    """Test copy module and its dependencies to empty repository.

    - Create repository and populate it
    - Create empty repository
    - Use 'copy' to copy 'nodejs' module with dependencies
    - assert module and its dependencies and relevant artifacts were copied
    """
    empty_repo = rpm_repository_factory()
    repo = rpm_modular_repo_on_demand

    modules = rpm_modulemd_api.list(
        repository_version=repo.latest_version_href,
        name="nodejs",
        stream="11",
        version="20180920144611",
    )
    module_to_copy = [modules.results[0].pulp_href]

    data = Copy(
        config=[
            {
                "source_repo_version": repo.latest_version_href,
                "dest_repo": empty_repo.pulp_href,
                "content": module_to_copy,
            }
        ],
        dependency_solving=True,
    )
    monitor_task(rpm_copy_api.copy_content(data).task)

    empty_repo = rpm_repository_api.read(empty_repo.pulp_href)
    modules = rpm_modulemd_api.list(repository_version=empty_repo.latest_version_href).results
    module_names = [module.name for module in modules]

    # assert that only 3 modules are copied (original and one dependency)
    assert len(modules) == 2
    # assert dependencies package names
    for dependency in ["nodejs", "postgresql"]:
        assert dependency in module_names


@pytest.mark.parallel
def test_advisory_copy_child_detection(
    init_and_sync,
    monitor_task,
    rpm_advisory_api,
    rpm_copy_api,
    rpm_modulemd_api,
    rpm_package_api,
    rpm_repository_api,
    rpm_repository_factory,
):
    """Test copy advisory and its direct package & module dependencies to empty repository.

    No recursive dependencies.

    - Create repository and populate it
    - Create empty repository
    - Use 'copy' to copy an advisory
    - assert advisory and its dependencies were copied
    """
    empty_repo = rpm_repository_factory()
    repo, _ = init_and_sync(url=RPM_MODULAR_FIXTURE_URL)

    test_advisory_href = get_all_content_hrefs(
        rpm_advisory_api, repository_version=repo.latest_version_href, id="FEDORA-2019-0329090518"
    )[0]
    content_to_copy = [test_advisory_href]

    data = Copy(
        config=[
            {
                "source_repo_version": repo.latest_version_href,
                "dest_repo": empty_repo.pulp_href,
                "content": content_to_copy,
            }
        ],
        dependency_solving=False,
    )
    monitor_task(rpm_copy_api.copy_content(data).task)

    empty_repo = rpm_repository_api.read(empty_repo.pulp_href)

    empty_repo_packages = [
        pkg.name
        for pkg in rpm_package_api.list(repository_version=empty_repo.latest_version_href).results
    ]
    empty_repo_advisories = [
        advisory.id
        for advisory in rpm_advisory_api.list(
            repository_version=empty_repo.latest_version_href
        ).results
    ]
    empty_repo_modules = [
        module.name
        for module in rpm_modulemd_api.list(
            repository_version=empty_repo.latest_version_href
        ).results
    ]

    # check the specific advisory was copied
    assert len(empty_repo_advisories) == 1
    # assert that all dependant packages were copied, the direct children of the advisory
    assert len(empty_repo_packages) == 2
    # assert dependencies package names
    for dependency in ["postgresql", "nodejs"]:
        assert dependency in empty_repo_packages
        assert dependency in empty_repo_modules


def get_all_content_hrefs(api, **kwargs):
    """Fetch all the content using the provided content API and query params.

    De-paginates the results.
    """
    content_list = []

    while True:
        content = api.list(**kwargs, offset=len(content_list))
        page = content.results
        content_list.extend([content.pulp_href for content in page])
        if not content.next:
            break

    return content_list



import json
import subprocess

from pulp_rpm.tests.functional.constants import (
    RPM_UNSIGNED_FIXTURE_URL,
    RPM_UNSIGNED_FIXTURE_SIZE,
    RPM_KICKSTART_FIXTURE_URL,
    RPM_KICKSTART_FIXTURE_SIZE,
)


def test_repo_size(init_and_sync, delete_orphans_pre, monitor_task, pulpcore_bindings):
    """Test that RPM repos correctly report their on-disk artifact sizes."""
    monitor_task(pulpcore_bindings.OrphansCleanupApi.cleanup({"orphan_protection_time": 0}).task)
    repo, _ = init_and_sync(url=RPM_UNSIGNED_FIXTURE_URL, policy="on_demand")

    cmd = (
        "pulpcore-manager",
        "repository-size",
        "--repositories",
        repo.pulp_href,
        "--include-on-demand",
    )
    run = subprocess.run(cmd, capture_output=True, check=True)
    out = json.loads(run.stdout)

    # Assert basic items of report and test on-demand sizing
    assert len(out) == 1
    report = out[0]
    assert report["name"] == repo.name
    assert report["href"] == repo.pulp_href
    assert report["disk-size"] == 0
    assert report["on-demand-size"] == RPM_UNSIGNED_FIXTURE_SIZE

    _, _ = init_and_sync(repository=repo, url=RPM_UNSIGNED_FIXTURE_URL, policy="immediate")
    run = subprocess.run(cmd, capture_output=True, check=True)
    report = json.loads(run.stdout)[0]
    assert report["disk-size"] == RPM_UNSIGNED_FIXTURE_SIZE
    assert report["on-demand-size"] == 0


def test_kickstart_repo_size(init_and_sync, delete_orphans_pre, monitor_task, pulpcore_bindings):
    """Test that kickstart RPM repos correctly report their on-disk artifact sizes."""
    monitor_task(pulpcore_bindings.OrphansCleanupApi.cleanup({"orphan_protection_time": 0}).task)
    repo, _ = init_and_sync(url=RPM_KICKSTART_FIXTURE_URL, policy="on_demand")

    cmd = (
        "pulpcore-manager",
        "repository-size",
        "--repositories",
        repo.pulp_href,
        "--include-on-demand",
    )
    run = subprocess.run(cmd, capture_output=True, check=True)
    out = json.loads(run.stdout)

    # Assert basic items of report and test on-demand sizing
    assert len(out) == 1
    report = out[0]
    assert report["name"] == repo.name
    assert report["href"] == repo.pulp_href
    assert report["disk-size"] == 2275  # One file is always downloaded
    assert report["on-demand-size"] == 133810  # Not all remote artifacts have sizes

    _, _ = init_and_sync(repository=repo, url=RPM_KICKSTART_FIXTURE_URL, policy="immediate")
    run = subprocess.run(cmd, capture_output=True, check=True)
    report = json.loads(run.stdout)[0]
    assert report["disk-size"] == RPM_KICKSTART_FIXTURE_SIZE
    assert report["on-demand-size"] == 0



"""Tests that CRUD rpm remotes."""

from random import choice
from uuid import uuid4

import pytest
from pulpcore.client.pulp_rpm.exceptions import ApiException

from pulp_rpm.tests.functional.constants import DOWNLOAD_POLICIES
from pulp_rpm.tests.functional.utils import gen_rpm_remote


@pytest.mark.parallel
def test_basic_crud_remote(rpm_rpmremote_api, rpm_rpmremote_factory, monitor_task):
    """Test CRUD operations for remotes."""
    # Create a remote
    body = _gen_verbose_remote()
    remote = rpm_rpmremote_factory(**body)
    for key in ("username", "password"):
        del body[key]

    for key, val in body.items():
        assert remote.to_dict()[key] == val, key

    # Try to create a second remote with an identical name
    body = gen_rpm_remote()
    body["name"] = remote.name
    with pytest.raises(ApiException):
        rpm_rpmremote_api.create(body)

    # Read a remote by its href
    remote = rpm_rpmremote_api.read(remote.pulp_href)
    for key, val in remote.to_dict().items():
        assert remote.to_dict()[key] == val, key

    # Read a remote by its name
    results = rpm_rpmremote_api.list(name=remote.name).results
    assert len(results) == 1
    for key, val in remote.to_dict().items():
        assert results[0].to_dict()[key] == val, key

    # Update a remote using HTTP PATCH
    body = _gen_verbose_remote()
    response = rpm_rpmremote_api.partial_update(remote.pulp_href, body)
    monitor_task(response.task)
    for key in ("username", "password"):
        del body[key]
    remote = rpm_rpmremote_api.read(remote.pulp_href)
    for key, val in body.items():
        assert remote.to_dict()[key] == val, key

    # Update a remote using HTTP PUT
    body = _gen_verbose_remote()
    response = rpm_rpmremote_api.update(remote.pulp_href, body)
    monitor_task(response.task)
    for key in ("username", "password"):
        del body[key]
    remote = rpm_rpmremote_api.read(remote.pulp_href)
    for key, val in body.items():
        assert remote.to_dict()[key] == val, key

    # Delete a remote
    response = rpm_rpmremote_api.delete(remote.pulp_href)
    monitor_task(response.task)
    with pytest.raises(ApiException):
        rpm_rpmremote_api.read(remote.pulp_href)


@pytest.mark.parallel
def test_missing_url(rpm_rpmremote_api):
    """Verify whether is possible to create a remote without a URL.

    This test targets the following issues:

    * `Pulp #3395 <https://pulp.plan.io/issues/3395>`_
    * `Pulp Smash #984 <https://github.com/pulp/pulp-smash/issues/984>`_
    """
    body = gen_rpm_remote()
    del body["url"]
    with pytest.raises(ApiException):
        rpm_rpmremote_api.create(body)


@pytest.mark.parallel
def test_policy_update_changes(rpm_rpmremote_api, rpm_rpmremote_factory, monitor_task):
    """Verify download policy behavior for valid and invalid values.

    In Pulp 3, there are different download policies.

    This test targets the following testing scenarios:

    1. Creating a remote without a download policy.
       Verify the creation is successful and immediate it is policy applied.
    2. Change the remote policy from default.
       Verify the change is successful.
    3. Attempt to change the remote policy to an invalid string.
       Verify an ApiException is given for the invalid policy as well
       as the policy remaining unchanged.

    For more information on the remote policies, see the Pulp3
    API on an installed server:

    * /pulp/api/v3/docs/#operation`

    This test targets the following issues:

    * `Pulp #4420 <https://pulp.plan.io/issues/4420>`_
    * `Pulp #3763 <https://pulp.plan.io/issues/3763>`_
    """
    # Verify the default policy `immediate`
    body = _gen_verbose_remote()
    del body["policy"]
    remote = rpm_rpmremote_factory(**body).to_dict()
    assert remote["policy"] == "immediate", remote

    # Verify ability to change policy to value other than the default
    changed_policy = choice([item for item in DOWNLOAD_POLICIES if item != "immediate"])
    response = rpm_rpmremote_api.partial_update(remote["pulp_href"], {"policy": changed_policy})
    monitor_task(response.task)

    remote = rpm_rpmremote_api.read(remote["pulp_href"]).to_dict()
    assert remote["policy"] == changed_policy, remote

    # Verify an invalid policy does not update the remote policy
    with pytest.raises(ApiException):
        rpm_rpmremote_api.partial_update(remote["pulp_href"], {"policy": str(uuid4())})


def test_raise_on_invalid_remote_url(
    rpm_rpmremote_api, rpm_ulnremote_api, gen_object_with_cleanup, monitor_task
):
    # Cant create invalid RpmRemote
    for protocol in ("uln", "sftp", "grpc"):
        with pytest.raises(ApiException, match=" is not valid. It must start with: "):
            body = _gen_verbose_remote()
            body["url"] = f"{protocol}://some/rpm/remote"
            gen_object_with_cleanup(rpm_rpmremote_api, body)

    # Cant create invalid UlnRemote
    for protocol in ("http", "https", "file"):
        with pytest.raises(ApiException, match=" is not valid. It must start with: "):
            body = _gen_verbose_remote()
            body["url"] = f"{protocol}://some/uln/remote"
            gen_object_with_cleanup(rpm_ulnremote_api, body)


def _gen_verbose_remote():
    """Return a semi-random dict for use in defining a remote.

    For most tests, it"s desirable to create remotes with as few attributes
    as possible, so that the tests can specifically target and attempt to break
    specific features. This module specifically targets remotes, so it makes
    sense to provide as many attributes as possible.

    Note that 'username' and 'password' are write-only attributes.
    """
    attrs = gen_rpm_remote()
    attrs.update(
        {"password": str(uuid4()), "username": str(uuid4()), "policy": choice(DOWNLOAD_POLICIES)}
    )
    return attrs



from unittest import TestCase
from datetime import datetime
from pulp_rpm.app.shared_utils import is_previous_version, urlpath_sanitize, parse_time


class TestSharedUtils(TestCase):
    """Test shared_utils functions."""

    def test_is_previous_version(self):
        """Test version-comparator."""
        # Versions must be int or 1.2.3
        # non-integer versions return False, always
        # True if version <= target

        # None
        self.assertTrue(is_previous_version(None, "1"))
        self.assertTrue(is_previous_version("1", None))
        self.assertTrue(is_previous_version(None, None))

        # Integer versions
        # v = t : v < t : v > t
        self.assertTrue(is_previous_version("1", "1"))
        self.assertTrue(is_previous_version("1", "2"))
        self.assertFalse(is_previous_version("2", "1"))

        # m.n
        # v = t : v m. < t m. : v m.n < t m.n : v m. > t.m : v m.n > t.m.n
        self.assertTrue(is_previous_version("1.2", "1.2"))
        self.assertTrue(is_previous_version("1.2", "2.2"))
        self.assertTrue(is_previous_version("1.2", "1.3"))
        self.assertFalse(is_previous_version("2.2", "1.2"))
        self.assertFalse(is_previous_version("2.2", "2.1"))

        # non-numeric : v not-digits : t not-digits : v-dot-nondigits : t dot-non-digits
        self.assertFalse(is_previous_version("foo", "1.2"))
        self.assertFalse(is_previous_version("1.2", "bar"))
        self.assertFalse(is_previous_version("foo.2", "2.1"))
        self.assertFalse(is_previous_version("1.2", "bar.1"))
        self.assertTrue(is_previous_version("1.foo", "2.bar"))
        self.assertFalse(is_previous_version("1.foo", "1.bar"))

    def test_urlpath_sanitize(self):
        """Test urljoin-replacement."""
        # arbitrary number of args become one single-slash-separated string
        a_expected = "a"
        ab_expected = "a/b"
        abc_expected = "a/b/c"

        # a /a a/ /a/
        self.assertEqual(a_expected, urlpath_sanitize("a"))
        self.assertEqual(a_expected, urlpath_sanitize("/a"))
        self.assertEqual(a_expected, urlpath_sanitize("a/"))
        self.assertEqual(a_expected, urlpath_sanitize("/a/"))

        # a b : a/ b : /a b : a b/ : a /b : a /b/ : a/ /b
        self.assertEqual(ab_expected, urlpath_sanitize("a", "b"))
        self.assertEqual(ab_expected, urlpath_sanitize("a/", "b"))
        self.assertEqual(ab_expected, urlpath_sanitize("/a", "b"))
        self.assertEqual(ab_expected, urlpath_sanitize("a", "b/"))
        self.assertEqual(ab_expected, urlpath_sanitize("a", "/b"))
        self.assertEqual(ab_expected, urlpath_sanitize("a", "/b/"))
        self.assertEqual(ab_expected, urlpath_sanitize("a/", "/b"))
        self.assertEqual(ab_expected, urlpath_sanitize("a/", "", "/b"))
        self.assertEqual(ab_expected, urlpath_sanitize("a/", "/", "/b"))

        # a b c : a /b/ /c : /a/ /b/ /c/
        self.assertEqual(abc_expected, urlpath_sanitize("a", "b", "c"))
        self.assertEqual(abc_expected, urlpath_sanitize("a", "/b/", "/c"))
        self.assertEqual(abc_expected, urlpath_sanitize("/a/", "/b/", "/c/"))

    def test_parse_time(self):
        """Test parse_time."""
        ts_input = "1626714399"
        ts_expected = 1626714399
        iso_input = "2012-01-27 16:08:06"
        iso_expected = datetime(2012, 1, 27, 16, 8, 6)

        self.assertEqual(ts_expected, parse_time(ts_input))
        self.assertNotEqual(ts_input, parse_time(ts_input))

        self.assertEqual(iso_expected, parse_time(iso_input))
        self.assertNotEqual(iso_input, parse_time(iso_input))

        self.assertIsNone(parse_time("abcd"))



# Generated by Django 4.2.5 on 2023-11-07 03:51

from django.db import migrations, models


def set_publication_checksum(apps, schema_editor):
    RpmPublication = apps.get_model("rpm", "RpmPublication")
    RpmPublication.objects.filter(checksum_type="").update(checksum_type="unknown")


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0059_rpmpublication_compression_type_and_more'),
    ]

    operations = [
        migrations.RunPython(set_publication_checksum),
    ]



# Generated by Django 4.2.5 on 2023-11-07 03:51

from django.db import migrations, models
from django.db.models import F


def set_publication_checksum(apps, schema_editor):
    RpmPublication = apps.get_model("rpm", "RpmPublication")
    RpmPublication.objects.update(checksum_type=F("metadata_checksum_type"))


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0056_remove_rpmpublication_sqlite_metadata_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='rpmpublication',
            name='checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True),
        ),
        migrations.RunPython(set_publication_checksum),
        migrations.AlterField(
            model_name='rpmpublication',
            name='checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')]),
        ),
        migrations.AddField(
            model_name='rpmrepository',
            name='checksum_type',
            field=models.TextField(choices=[('unknown', 'unknown'), ('md5', 'md5'), ('sha1', 'sha1'), ('sha1', 'sha1'), ('sha224', 'sha224'), ('sha256', 'sha256'), ('sha384', 'sha384'), ('sha512', 'sha512')], null=True),
        ),
    ]



from pulpcore.plugin.viewsets import ReadOnlyContentViewSet

from pulp_rpm.app.models import RepoMetadataFile
from pulp_rpm.app.serializers import RepoMetadataFileSerializer


class RepoMetadataFileViewSet(ReadOnlyContentViewSet):
    """
    RepoMetadataFile Viewset.
    """

    endpoint_name = "repo_metadata_files"
    queryset = RepoMetadataFile.objects.all()
    serializer_class = RepoMetadataFileSerializer

    DEFAULT_ACCESS_POLICY = {
        "statements": [
            {
                "action": ["list", "retrieve"],
                "principal": "authenticated",
                "effect": "allow",
            },
        ],
        "queryset_scoping": {"function": "scope_queryset"},
    }



"""Tests that perform actions over advisory content unit upload."""

import pytest
import os

from pulpcore.tests.functional.utils import PulpTaskError
from pulp_rpm.tests.functional.constants import (
    BEAR_JSON,
    CAMEL_BEAR_DOG_JSON,
    CAMEL_BIRD_JSON,
    CAMEL_JSON,
    CESA_2020_5002,
    CESA_2020_4910,
    RPM_PACKAGE_FILENAME,
    RPM_UNSIGNED_FIXTURE_URL,
)


@pytest.mark.parallel
def test_upload_wrong_type(upload_wrong_file_type, monitor_task):
    """Test that a proper error is raised when wrong file content type is uploaded."""
    bad_file_to_use = os.path.join(RPM_UNSIGNED_FIXTURE_URL, RPM_PACKAGE_FILENAME)
    with pytest.raises(PulpTaskError) as e:
        monitor_task(upload_wrong_file_type(bad_file_to_use).task)
    assert "JSON" in e.value.task.error["description"]


@pytest.mark.parallel
def test_upload_json(upload_advisory_factory):
    """Test upload advisory from JSON file."""
    advisory1, _, an_id = upload_advisory_factory(set_id=True)
    assert advisory1.id == an_id


@pytest.mark.parallel
def test_merging(
    upload_advisory_factory,
    assert_uploaded_advisory,
    rpm_advisory_api,
    rpm_repository_factory,
):
    """Test the 'same' advisory, diff pkglists, into a repo, expecting a merged package-list."""
    repo = rpm_repository_factory()
    bear, vers_href, an_id = upload_advisory_factory(
        advisory=BEAR_JSON, repository=repo, set_id=True
    )
    bear = rpm_advisory_api.read(bear.pulp_href)
    assert_uploaded_advisory(an_id, vers_href)
    assert vers_href == f"{repo.pulp_href}versions/1/"
    assert 1 == len(bear.pkglist)
    assert 1 == len(bear.pkglist[0].packages)

    # Second upload, no pkg-intersection - add both collections
    # NOTE: also check that unnamed-collections are now named "collection_N", so
    # they can be uniquely identified
    _, vers_href, _ = upload_advisory_factory(advisory=CAMEL_JSON, repository=repo, use_id=an_id)
    advisory_href, vers_herf = assert_uploaded_advisory(an_id, vers_href)
    assert vers_href == f"{repo.pulp_href}versions/2/"
    cambear = rpm_advisory_api.read(advisory_href)
    assert 2 == len(cambear.pkglist)
    coll_names = [row.name for row in cambear.pkglist]
    assert "collection_0" in coll_names
    assert "collection_1" in coll_names
    assert 1 == len(cambear.pkglist[0].packages)
    assert 1 == len(cambear.pkglist[1].packages)
    names = [plist.packages[0]["name"] for plist in cambear.pkglist]
    assert "camel" in names
    assert "bear" in names

    # Third upload, two pkgs, intersects with existing, expect AdvisoryConflict failure
    with pytest.raises(PulpTaskError) as ctx:
        _, _, _ = upload_advisory_factory(advisory=CAMEL_BIRD_JSON, repository=repo, use_id=an_id)
    assert "neither package list is a proper subset of the other" in str(ctx.value)
    assert "ALLOW_AUTOMATIC_UNSAFE_ADVISORY_CONFLICT_RESOLUTION" in str(ctx.value)

    # Fourth upload, intersecting pkglists, expecting three pkgs
    cambeardog, vers_href, _ = upload_advisory_factory(
        advisory=CAMEL_BEAR_DOG_JSON, repository=repo, use_id=an_id
    )
    assert_uploaded_advisory(an_id, vers_href)
    assert vers_href == f"{repo.pulp_href}versions/3/"
    assert an_id == cambeardog.id
    assert 1 == len(cambeardog.pkglist)
    # Expect one collection, not a merge
    names = [pkg["name"] for pkg in cambeardog.pkglist[0].packages]
    assert 3 == len(names)
    assert "camel" in names
    assert "bear" in names
    assert "dog" in names


@pytest.mark.parallel
def test_8683_error_path(
    upload_advisory_factory,
    assert_uploaded_advisory,
    rpm_repository_factory,
):
    """
    Test that upload-fail doesn't break all future uploads.

    See https://pulp.plan.io/issues/8683 for details.
    """
    # Upload an advisory
    repo = rpm_repository_factory()

    _, vers_href, id1 = upload_advisory_factory(
        advisory=CESA_2020_5002, repository=repo, set_id=True
    )
    assert_uploaded_advisory(id1, vers_href)

    # Try to upload it 'again' and watch it fail
    with pytest.raises(PulpTaskError):
        _, _, _ = upload_advisory_factory(advisory=CESA_2020_5002, repository=repo, use_id=id1)

    # Upload a different advisory and Don't Fail
    advisory3, _, id2 = upload_advisory_factory(
        advisory=CESA_2020_4910, repository=repo, set_id=True
    )
    assert id2 == advisory3.id
